"""
MCP 服务器

包含 10 个只读查询工具：
0. get_devices - 列出所有设备
1. get_status - 设备实时状态
2. get_dnd - 勿扰模式设置
3. get_clean_summary - 清洁历史统计
4. get_sound_volume - 音量设置
5. get_rooms - 房间列表
6. get_maps - 地图概览
7. get_map_content - 地图内容
8. get_consumables - 耗材使用状态
9. get_home - 家庭完整布局
"""

from typing import Any

from mcp.server.fastmcp import FastMCP

from .connection import ConnectionManager, lifespan

# 创建 FastMCP 实例
mcp = FastMCP(
    "roborock-mcp",
    lifespan=lifespan,
)


def get_conn() -> ConnectionManager:
    """从 MCP context 获取 ConnectionManager"""
    return mcp.get_context().request_context.lifespan_context["conn"]


# ============================================================================
# 工具定义
# ============================================================================

@mcp.tool()
async def get_devices() -> dict[str, Any]:
    """
    列出所有已发现的设备

    返回设备的名称、型号、DUID 和连接状态。
    """
    conn = get_conn()
    devices = await conn.list_devices()

    result = {
        "count": len(devices),
        "devices": []
    }

    for device in devices:
        info = {
            "name": device.name,
            "duid": device.duid,
            "model": device.product.model if device.product else None,
            "firmware": device.device_info.fv if device.device_info else None,
        }
        result["devices"].append(info)

    return result


@mcp.tool()
async def get_status(device_name: str | None = None) -> dict[str, Any]:
    """
    获取设备实时状态

    包括电量、状态、吸力、拖布湿度、基座信息、错误码等。

    Args:
        device_name: 设备名称（可选，单设备时自动选择）

    Returns:
        设备状态信息，包含 state（charging/cleaning/idle/paused/error 等）、
        battery（电量百分比）、fan_speed（吸力档位）、water_mode（拖布湿度）等
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.status.refresh()
    status = props.status

    return {
        "device_name": device.name,
        "state": status.state.name if status.state else None,
        "state_name": status.state_name,
        "battery": status.battery,
        "error_code": status.error_code.name if status.error_code else None,
        "error_name": status.error_code_name,
        "fan_speed": status.fan_speed_name,
        "water_mode": status.water_mode_name,
        "mop_route": status.mop_route_name,
        "clean_time_seconds": status.clean_time,
        "clean_area_m2": status.square_meter_clean_area,
        "clean_percent": status.clean_percent,
        "dock_type": status.dock_type.name if status.dock_type else None,
        "dock_error": status.dock_error_status.name if status.dock_error_status else None,
        "charge_status": status.charge_status,
        "dnd_enabled": bool(status.dnd_enabled),
        "map_present": bool(status.map_present),
    }


@mcp.tool()
async def get_dnd(device_name: str | None = None) -> dict[str, Any]:
    """
    获取勿扰模式设置

    包括是否开启、开始和结束时间。

    Args:
        device_name: 设备名称（可选，单设备时自动选择）
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.dnd.refresh()
    dnd = props.dnd

    return {
        "device_name": device.name,
        "enabled": dnd.is_on,
        "start_time": f"{dnd.start_hour:02d}:{dnd.start_minute:02d}" if dnd.is_on else None,
        "end_time": f"{dnd.end_hour:02d}:{dnd.end_minute:02d}" if dnd.is_on else None,
    }


@mcp.tool()
async def get_clean_summary(device_name: str | None = None) -> dict[str, Any]:
    """
    获取清洁历史统计

    包括总清洁时长、面积、次数，以及最近一次清洁记录。

    Args:
        device_name: 设备名称（可选，单设备时自动选择）
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.clean_summary.refresh()
    summary = props.clean_summary

    result = {
        "device_name": device.name,
        "total_time_hours": summary.clean_time / 3600 if summary.clean_time else 0,
        "total_area_m2": summary.clean_area / 1_000_000 if summary.clean_area else 0,
        "clean_count": summary.clean_count,
        "record_count": len(summary.records) if summary.records else 0,
    }

    # 最近一次清洁记录
    if summary.last_clean_record:
        record = summary.last_clean_record
        result["last_clean"] = {
            "begin": record.begin_datetime.isoformat() if record.begin_datetime else None,
            "end": record.end_datetime.isoformat() if record.end_datetime else None,
            "duration_seconds": record.duration,
            "area_m2": record.area / 1_000_000 if record.area else 0,
            "start_type": record.start_type.name if record.start_type else None,
            "avoid_count": record.avoid_count,
        }

    return result


@mcp.tool()
async def get_sound_volume(device_name: str | None = None) -> dict[str, Any]:
    """
    获取音量设置

    Args:
        device_name: 设备名称（可选，单设备时自动选择）
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.sound_volume.refresh()
    volume = props.sound_volume

    return {
        "device_name": device.name,
        "volume": volume.volume,
    }


@mcp.tool()
async def get_rooms(device_name: str | None = None) -> dict[str, Any]:
    """
    获取房间列表

    返回房间 ID（segment_id）与名称的映射。

    Args:
        device_name: 设备名称（可选，单设备时自动选择）
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.rooms.refresh()
    rooms = props.rooms

    room_list = []
    for room in rooms.rooms or []:
        room_list.append({
            "segment_id": room.segment_id,
            "name": room.raw_name or "<未命名>",
        })

    return {
        "device_name": device.name,
        "room_count": len(room_list),
        "rooms": room_list,
    }


@mcp.tool()
async def get_maps(device_name: str | None = None) -> dict[str, Any]:
    """
    获取地图概览

    轻量查询，返回地图列表、当前地图 ID 和房间数。
    如需完整的房间布局，请使用 get_home。

    Args:
        device_name: 设备名称（可选，单设备时自动选择）
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.maps.refresh()
    maps = props.maps

    map_list = []
    for map_info in maps.map_info or []:
        map_list.append({
            "map_flag": map_info.map_flag,
            "name": map_info.name or "<未命名>",
            "room_count": len(map_info.rooms) if map_info.rooms else 0,
            "is_current": map_info.map_flag == maps.current_map,
        })

    return {
        "device_name": device.name,
        "current_map": maps.current_map,
        "map_count": len(map_list),
        "maps": map_list,
    }


@mcp.tool()
async def get_map_content(device_name: str | None = None) -> dict[str, Any]:
    """
    获取地图内容

    返回地图元数据（图片大小、机器人位置、充电座位置）。
    不返回原始图片字节。

    Args:
        device_name: 设备名称（可选，单设备时自动选择）
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.map_content.refresh()
    content = props.map_content

    result = {
        "device_name": device.name,
        "image_size_bytes": len(content.image_content) if content.image_content else 0,
        "has_map_data": content.map_data is not None,
    }

    if content.map_data:
        vacuum_pos = content.map_data.vacuum_position
        charger_pos = content.map_data.charger
        result["vacuum_position"] = {
            "x": vacuum_pos.x if vacuum_pos else None,
            "y": vacuum_pos.y if vacuum_pos else None,
        }
        result["charger_position"] = {
            "x": charger_pos.x if charger_pos else None,
            "y": charger_pos.y if charger_pos else None,
        }

    return result


@mcp.tool()
async def get_consumables(device_name: str | None = None) -> dict[str, Any]:
    """
    获取耗材使用状态

    包括主刷、边刷、滤网、传感器的工作时长（小时）。

    Args:
        device_name: 设备名称（可选，单设备时自动选择）
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties
    await props.consumables.refresh()
    consumables = props.consumables

    # 工作时长从分钟转换为小时
    HOURS = 60

    return {
        "device_name": device.name,
        "main_brush_hours": consumables.main_brush_work_time / HOURS if consumables.main_brush_work_time else 0,
        "side_brush_hours": consumables.side_brush_work_time / HOURS if consumables.side_brush_work_time else 0,
        "filter_hours": consumables.filter_work_time / HOURS if consumables.filter_work_time else 0,
        "sensor_hours": consumables.sensor_dirty_time / HOURS if consumables.sensor_dirty_time else 0,
    }


@mcp.tool()
async def get_home(device_name: str | None = None) -> dict[str, Any]:
    """
    获取家庭完整布局

    重量查询，返回所有地图的完整房间布局。
    SDK 内部已处理设备忙时的降级逻辑。

    Args:
        device_name: 设备名称（可选，单设备时自动选择）
    """
    conn = get_conn()
    device = await conn.get_device(device_name)

    if not device.v1_properties:
        return {"error": "设备不支持 V1 协议"}

    props = device.v1_properties

    # 直接调用 refresh()，SDK 内部已处理设备忙的降级逻辑
    await props.home.refresh()
    home = props.home

    maps_info = {}
    for map_flag, map_info in (home.home_map_info or {}).items():
        rooms = []
        for room in map_info.rooms or []:
            rooms.append({
                "segment_id": room.segment_id,
                "name": room.raw_name or "<未命名>",
            })
        maps_info[str(map_flag)] = {
            "name": map_info.name or "<未命名>",
            "rooms": rooms,
        }

    # 当前地图的房间
    current_rooms = []
    for room in home.current_rooms or []:
        current_rooms.append({
            "segment_id": room.segment_id,
            "name": room.raw_name or "<未命名>",
        })

    return {
        "device_name": device.name,
        "map_count": len(maps_info),
        "maps": maps_info,
        "current_rooms": current_rooms,
    }



from dataclasses import Field, fields

from roborock.data import AppInitStatus, HomeDataProduct, RoborockBase
from roborock.data.v1.v1_containers import FieldNameBase
from roborock.device_features import DeviceFeatures
from roborock.devices.cache import DeviceCache
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand


class DeviceFeaturesTrait(DeviceFeatures, common.V1TraitMixin):
    """Trait for managing supported features on Roborock devices."""

    command = RoborockCommand.APP_GET_INIT_STATUS

    def __init__(self, product: HomeDataProduct, device_cache: DeviceCache) -> None:  # pylint: disable=super-init-not-called
        """Initialize DeviceFeaturesTrait."""
        self._product = product
        self._nickname = product.product_nickname
        self._device_cache = device_cache
        # All fields of DeviceFeatures are required. Initialize them to False
        # so we have some known state.
        for field in fields(self):
            setattr(self, field.name, False)

    def is_field_supported(self, cls: type[RoborockBase], field_name: FieldNameBase) -> bool:
        """Determines if the specified field is supported by this device.

        We use dataclass attributes on the field to specify the schema code that is required
        for the field to be supported and it is compared against the list of
        supported schema codes for the device returned in the product information.
        """
        dataclass_field: Field | None = None
        for field in fields(cls):
            if field.name == field_name:
                dataclass_field = field
                break
        if dataclass_field is None:
            raise ValueError(f"Field {field_name} not found in {cls}")

        requires_schema_code = dataclass_field.metadata.get("requires_schema_code", None)
        if requires_schema_code is None:
            # We assume the field is supported
            return True
        # If the field requires a protocol that is not supported, we return False
        return requires_schema_code in self._product.supported_schema_codes

    async def refresh(self) -> None:
        """Refresh the contents of this trait.

        This will use cached device features if available since they do not
        change often and this avoids unnecessary RPC calls. This would only
        ever change with a firmware update, so caching is appropriate.
        """
        cache_data = await self._device_cache.get()
        if cache_data.device_features is not None:
            self._update_trait_values(cache_data.device_features)
            return
        # Save cached device features
        await super().refresh()
        cache_data.device_features = self
        await self._device_cache.set(cache_data)

    def _parse_response(self, response: common.V1ResponseData) -> DeviceFeatures:
        """Parse the response from the device into a MapContentTrait instance."""
        if not isinstance(response, list):
            raise ValueError(f"Unexpected AppInitStatus response format: {type(response)}")
        app_status = AppInitStatus.from_dict(response[0])
        return DeviceFeatures.from_feature_flags(
            new_feature_info=app_status.new_feature_info,
            new_feature_info_str=app_status.new_feature_info_str,
            feature_info=app_status.feature_info,
            product_nickname=self._nickname,
        )



"""An MQTT session for sending and receiving messages."""

from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass, field

from roborock.diagnostics import Diagnostics
from roborock.exceptions import RoborockException
from roborock.mqtt.health_manager import HealthManager

DEFAULT_TIMEOUT = 30.0

SessionUnauthorizedHook = Callable[[], None]


@dataclass
class MqttParams:
    """MQTT parameters for the connection."""

    host: str
    """MQTT host to connect to."""

    port: int
    """MQTT port to connect to."""

    tls: bool
    """Use TLS for the connection."""

    username: str
    """MQTT username to use for authentication."""

    password: str
    """MQTT password to use for authentication."""

    verify_tls: bool = True
    """Verify the TLS certificate."""

    timeout: float = DEFAULT_TIMEOUT
    """Timeout for communications with the broker in seconds."""

    diagnostics: Diagnostics = field(default_factory=Diagnostics)
    """Diagnostics object for tracking MQTT session stats.

    This defaults to a new Diagnostics object, but the common case is the
    caller will provide their own (e.g., from a DeviceManager) so that the
    shared MQTT session diagnostics are included in the overall diagnostics.
    """

    unauthorized_hook: SessionUnauthorizedHook | None = None
    """Optional hook invoked when an unauthorized error is received.

    This may be invoked by the background reconnect logic when an
    unauthorized error is received from the broker. The caller may use
    this hook to refresh credentials or take other actions as needed.
    """


class MqttSession(ABC):
    """An MQTT session for sending and receiving messages."""

    @property
    @abstractmethod
    def connected(self) -> bool:
        """True if the session is connected to the broker."""

    @property
    @abstractmethod
    def health_manager(self) -> HealthManager:
        """Return the health manager for the session."""

    @abstractmethod
    async def subscribe(self, device_id: str, callback: Callable[[bytes], None]) -> Callable[[], None]:
        """Invoke the callback when messages are received on the topic.

        The returned callable unsubscribes from the topic when called.
        """

    @abstractmethod
    async def publish(self, topic: str, message: bytes) -> None:
        """Publish a message on the specified topic.

        This will raise an exception if the message could not be sent.
        """

    @abstractmethod
    async def restart(self) -> None:
        """Force the session to disconnect and reconnect."""

    @abstractmethod
    async def close(self) -> None:
        """Cancels the mqtt loop"""


class MqttSessionException(RoborockException):
    """Raised when there is an error communicating with MQTT."""


class MqttSessionUnauthorized(RoborockException):
    """Raised when there is an authorization error communicating with MQTT.

    This error may be raised in multiple scenarios so there is not a well
    defined behavior for how the caller should behave. The two cases are:
    - Rate limiting is in effect and the caller should retry after some time.
    - The credentials are invalid and the caller needs to obtain new credentials

    However, it is observed that obtaining new credentials may resolve the
    issue in both cases.
    """



from roborock.data import DnDTimer
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand

_ENABLED_PARAM = "enabled"


class DoNotDisturbTrait(DnDTimer, common.V1TraitMixin, common.RoborockSwitchBase):
    """Trait for managing Do Not Disturb (DND) settings on Roborock devices."""

    command = RoborockCommand.GET_DND_TIMER

    @property
    def is_on(self) -> bool:
        """Return whether the Do Not Disturb (DND) timer is enabled."""
        return self.enabled == 1

    async def set_dnd_timer(self, dnd_timer: DnDTimer) -> None:
        """Set the Do Not Disturb (DND) timer settings of the device."""
        await self.rpc_channel.send_command(RoborockCommand.SET_DND_TIMER, params=dnd_timer.as_list())
        await self.refresh()

    async def clear_dnd_timer(self) -> None:
        """Clear the Do Not Disturb (DND) timer settings of the device."""
        await self.rpc_channel.send_command(RoborockCommand.CLOSE_DND_TIMER)
        await self.refresh()

    async def enable(self) -> None:
        """Set the Do Not Disturb (DND) timer settings of the device."""
        await self.rpc_channel.send_command(
            RoborockCommand.SET_DND_TIMER,
            params=self.as_list(),
        )
        # Optimistic update to avoid an extra refresh
        self.enabled = 1

    async def disable(self) -> None:
        """Disable the Do Not Disturb (DND) timer settings of the device."""
        await self.rpc_channel.send_command(RoborockCommand.CLOSE_DND_TIMER)
        # Optimistic update to avoid an extra refresh
        self.enabled = 0



"""Trait for wash towel mode."""

from roborock.data import WashTowelMode
from roborock.device_features import is_wash_n_fill_dock
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand


class WashTowelModeTrait(WashTowelMode, common.V1TraitMixin):
    """Trait for wash towel mode."""

    command = RoborockCommand.GET_WASH_TOWEL_MODE
    requires_dock_type = is_wash_n_fill_dock



"""Common utilities for Q10 traits.

This module provides infrastructure for mapping Roborock Data Points (DPS) to
Python dataclass fields and handling the lifecycle of data updates from the
device.

### DPS Metadata Annotation

Classes extending `RoborockBase` can annotate their fields with DPS IDs using
the `field(metadata={"dps": ...})` convention. This creates a declarative
mapping that `DpsDataConverter` uses to automatically route incoming device
data to the correct attribute.

Example:

```python
@dataclass
class MyStatus(RoborockBase):
    battery: int = field(metadata={"dps": B01_Q10_DP.BATTERY})
```

### Update Lifecycle
1. **Raw Data**: The device sends encoded DPS updates over MQTT.
2. **Decoding**: The transport layer decodes these into a dictionary (e.g., `{"101": 80}`).
3. **Conversion**: `DpsDataConverter` uses `RoborockBase.convert_dict` to transform
   raw values into appropriate Python types (e.g., Enums, ints) based on the
   dataclass field types.
4. **Update**: `update_from_dps` maps these converted values to field names and
   updates the target object using `setattr`.

### Usage

Typically, a trait will instantiate a single `DpsDataConverter` for its status class
and call `update_from_dps` whenever new data is received from the device stream.

"""

import dataclasses
import logging
from collections.abc import Callable
from typing import Any

from roborock.callbacks import CallbackList
from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP
from roborock.data.containers import RoborockBase


class TraitUpdateListener:
    """Trait update listener.

    This is a base class for traits to support notifying listeners when they
    have been updated. Clients may register callbacks to be notified when the
    trait has been updated. When the listener callback is invoked, the client
    should read the trait's properties to get the updated values.
    """

    def __init__(self, logger: logging.Logger) -> None:
        """Initialize the trait update listener."""
        self._update_callbacks: CallbackList[None] = CallbackList(logger=logger)

    def add_update_listener(self, callback: Callable[[], None]) -> Callable[[], None]:
        """Register a callback when the trait has been updated.

        Returns a callable to remove the listener.
        """
        # We wrap the callback to ignore the value passed to it.
        return self._update_callbacks.add_callback(lambda _: callback())

    def _notify_update(self) -> None:
        """Notify all update listeners."""
        self._update_callbacks(None)


class DpsDataConverter:
    """Utility to handle the transformation and merging of DPS data into models.

    This class pre-calculates the mapping between Data Point IDs and dataclass fields
    to optimize repeated updates from device streams.
    """

    def __init__(self, dps_type_map: dict[B01_Q10_DP, type], dps_field_map: dict[B01_Q10_DP, str]):
        """Initialize the converter for a specific RoborockBase-derived class."""
        self._dps_type_map = dps_type_map
        self._dps_field_map = dps_field_map

    @classmethod
    def from_dataclass(cls, dataclass_type: type[RoborockBase]):
        """Initialize the converter for a specific RoborockBase-derived class."""
        dps_type_map: dict[B01_Q10_DP, type] = {}
        dps_field_map: dict[B01_Q10_DP, str] = {}
        for field_obj in dataclasses.fields(dataclass_type):
            if field_obj.metadata and "dps" in field_obj.metadata:
                dps_id = field_obj.metadata["dps"]
                dps_type_map[dps_id] = field_obj.type
                dps_field_map[dps_id] = field_obj.name
        return cls(dps_type_map, dps_field_map)

    def update_from_dps(self, target: RoborockBase, decoded_dps: dict[B01_Q10_DP, Any]) -> bool:
        """Convert and merge raw DPS data into the target object.

        Uses the pre-calculated type mapping to ensure values are converted to the
        correct Python types before being updated on the target.

        Args:
            target: The target object to update.
            decoded_dps: The decoded DPS data to convert.

        Returns:
            True if any values were updated, False otherwise.
        """
        conversions = RoborockBase.convert_dict(self._dps_type_map, decoded_dps)
        for dps_id, value in conversions.items():
            field_name = self._dps_field_map[dps_id]
            setattr(target, field_name, value)
        return bool(conversions)



import datetime
import logging
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Any

from roborock.const import (
    CLEANING_BRUSH_REPLACE_TIME,
    DUST_COLLECTION_REPLACE_TIME,
    FILTER_REPLACE_TIME,
    MAIN_BRUSH_REPLACE_TIME,
    MOP_ROLLER_REPLACE_TIME,
    NO_MAP,
    ROBOROCK_G10S_PRO,
    ROBOROCK_P10,
    ROBOROCK_Q7_MAX,
    ROBOROCK_QREVO_CURV,
    ROBOROCK_QREVO_MASTER,
    ROBOROCK_QREVO_MAXV,
    ROBOROCK_QREVO_PRO,
    ROBOROCK_QREVO_S,
    ROBOROCK_S4_MAX,
    ROBOROCK_S5_MAX,
    ROBOROCK_S6,
    ROBOROCK_S6_MAXV,
    ROBOROCK_S6_PURE,
    ROBOROCK_S7,
    ROBOROCK_S7_MAXV,
    ROBOROCK_S8,
    ROBOROCK_S8_MAXV_ULTRA,
    ROBOROCK_S8_PRO_ULTRA,
    ROBOROCK_SAROS_10,
    ROBOROCK_SAROS_10R,
    SENSOR_DIRTY_REPLACE_TIME,
    SIDE_BRUSH_REPLACE_TIME,
    STRAINER_REPLACE_TIME,
    ROBOROCK_G20S_Ultra,
)
from roborock.exceptions import RoborockException

from ..containers import NamedRoomMapping, RoborockBase, RoborockBaseTimer, _attr_repr
from .v1_code_mappings import (
    CleanFluidStatus,
    ClearWaterBoxStatus,
    DirtyWaterBoxStatus,
    DustBagStatus,
    RoborockCleanType,
    RoborockDockDustCollectionModeCode,
    RoborockDockErrorCode,
    RoborockDockTypeCode,
    RoborockDockWashTowelModeCode,
    RoborockErrorCode,
    RoborockFanPowerCode,
    RoborockFanSpeedP10,
    RoborockFanSpeedQ7Max,
    RoborockFanSpeedQRevoCurv,
    RoborockFanSpeedQRevoMaster,
    RoborockFanSpeedQRevoMaxV,
    RoborockFanSpeedS6Pure,
    RoborockFanSpeedS7,
    RoborockFanSpeedS7MaxV,
    RoborockFanSpeedS8MaxVUltra,
    RoborockFanSpeedSaros10,
    RoborockFanSpeedSaros10R,
    RoborockFinishReason,
    RoborockInCleaning,
    RoborockMopIntensityCode,
    RoborockMopIntensityP10,
    RoborockMopIntensityQ7Max,
    RoborockMopIntensityQRevoCurv,
    RoborockMopIntensityQRevoMaster,
    RoborockMopIntensityQRevoMaxV,
    RoborockMopIntensityS5Max,
    RoborockMopIntensityS6MaxV,
    RoborockMopIntensityS7,
    RoborockMopIntensityS8MaxVUltra,
    RoborockMopIntensitySaros10,
    RoborockMopIntensitySaros10R,
    RoborockMopModeCode,
    RoborockMopModeQRevoCurv,
    RoborockMopModeQRevoMaster,
    RoborockMopModeQRevoMaxV,
    RoborockMopModeS7,
    RoborockMopModeS8MaxVUltra,
    RoborockMopModeS8ProUltra,
    RoborockMopModeSaros10,
    RoborockMopModeSaros10R,
    RoborockStartType,
    RoborockStateCode,
)

_LOGGER = logging.getLogger(__name__)


class FieldNameBase(StrEnum):
    """A base enum class that represents a field name in a RoborockBase dataclass."""


class StatusField(FieldNameBase):
    """An enum that represents a field in the `Status` class.

    This is used with `roborock.devices.traits.v1.status.DeviceFeaturesTrait`
    to understand if a feature is supported by the device using `is_field_supported`.

    The enum values are names of fields in the `Status` class. Each field is
    annotated with `requires_schema_code` metadata to map the field to a schema
    code in the product schema, which may have a different name than the field/attribute name.
    """

    STATE = "state"
    BATTERY = "battery"
    FAN_POWER = "fan_power"
    WATER_BOX_MODE = "water_box_mode"
    CHARGE_STATUS = "charge_status"
    DRY_STATUS = "dry_status"


def _requires_schema_code(requires_schema_code: str, default=None) -> Any:
    return field(metadata={"requires_schema_code": requires_schema_code}, default=default)


@dataclass
class Status(RoborockBase):
    """This status will be deprecated in favor of StatusV2."""

    msg_ver: int | None = None
    msg_seq: int | None = None
    state: RoborockStateCode | None = _requires_schema_code("state", default=None)
    battery: int | None = _requires_schema_code("battery", default=None)
    clean_time: int | None = None
    clean_area: int | None = None
    error_code: RoborockErrorCode | None = None
    map_present: int | None = None
    in_cleaning: RoborockInCleaning | None = None
    in_returning: int | None = None
    in_fresh_state: int | None = None
    lab_status: int | None = None
    water_box_status: int | None = None
    back_type: int | None = None
    wash_phase: int | None = None
    wash_ready: int | None = None
    fan_power: RoborockFanPowerCode | None = _requires_schema_code("fan_power", default=None)
    dnd_enabled: int | None = None
    map_status: int | None = None
    is_locating: int | None = None
    lock_status: int | None = None
    water_box_mode: RoborockMopIntensityCode | None = _requires_schema_code("water_box_mode", default=None)
    water_box_carriage_status: int | None = None
    mop_forbidden_enable: int | None = None
    camera_status: int | None = None
    is_exploring: int | None = None
    home_sec_status: int | None = None
    home_sec_enable_password: int | None = None
    adbumper_status: list[int] | None = None
    water_shortage_status: int | None = None
    dock_type: RoborockDockTypeCode | None = None
    dust_collection_status: int | None = None
    auto_dust_collection: int | None = None
    avoid_count: int | None = None
    mop_mode: RoborockMopModeCode | None = None
    debug_mode: int | None = None
    collision_avoid_status: int | None = None
    switch_map_mode: int | None = None
    dock_error_status: RoborockDockErrorCode | None = None
    charge_status: int | None = _requires_schema_code("charge_status", default=None)
    unsave_map_reason: int | None = None
    unsave_map_flag: int | None = None
    wash_status: int | None = None
    distance_off: int | None = None
    in_warmup: int | None = None
    dry_status: int | None = _requires_schema_code("drying_status", default=None)
    rdt: int | None = None
    clean_percent: int | None = None
    rss: int | None = None
    dss: int | None = None
    common_status: int | None = None
    corner_clean_mode: int | None = None
    last_clean_t: int | None = None
    replenish_mode: int | None = None
    repeat: int | None = None
    kct: int | None = None
    subdivision_sets: int | None = None

    @property
    def square_meter_clean_area(self) -> float | None:
        return round(self.clean_area / 1000000, 1) if self.clean_area is not None else None

    @property
    def error_code_name(self) -> str | None:
        return self.error_code.name if self.error_code is not None else None

    @property
    def state_name(self) -> str | None:
        return self.state.name if self.state is not None else None

    @property
    def water_box_mode_name(self) -> str | None:
        return self.water_box_mode.name if self.water_box_mode is not None else None

    @property
    def fan_power_options(self) -> list[str]:
        if self.fan_power is None:
            return []
        return list(self.fan_power.keys())

    @property
    def fan_power_name(self) -> str | None:
        return self.fan_power.name if self.fan_power is not None else None

    @property
    def mop_mode_name(self) -> str | None:
        return self.mop_mode.name if self.mop_mode is not None else None

    def get_fan_speed_code(self, fan_speed: str) -> int:
        if self.fan_power is None:
            raise RoborockException("Attempted to get fan speed before status has been updated.")
        return self.fan_power.as_dict().get(fan_speed)

    def get_mop_intensity_code(self, mop_intensity: str) -> int:
        if self.water_box_mode is None:
            raise RoborockException("Attempted to get mop_intensity before status has been updated.")
        return self.water_box_mode.as_dict().get(mop_intensity)

    def get_mop_mode_code(self, mop_mode: str) -> int:
        if self.mop_mode is None:
            raise RoborockException("Attempted to get mop_mode before status has been updated.")
        return self.mop_mode.as_dict().get(mop_mode)

    @property
    def current_map(self) -> int | None:
        """Returns the current map ID if the map is present."""
        if self.map_status is not None:
            map_flag = self.map_status >> 2
            if map_flag != NO_MAP:
                return map_flag
        return None

    @property
    def clear_water_box_status(self) -> ClearWaterBoxStatus | None:
        if self.dss:
            return ClearWaterBoxStatus((self.dss >> 2) & 3)
        return None

    @property
    def dirty_water_box_status(self) -> DirtyWaterBoxStatus | None:
        if self.dss:
            return DirtyWaterBoxStatus((self.dss >> 4) & 3)
        return None

    @property
    def dust_bag_status(self) -> DustBagStatus | None:
        if self.dss:
            return DustBagStatus((self.dss >> 6) & 3)
        return None

    @property
    def water_box_filter_status(self) -> int | None:
        if self.dss:
            return (self.dss >> 8) & 3
        return None

    @property
    def clean_fluid_status(self) -> CleanFluidStatus | None:
        if self.dss:
            value = (self.dss >> 10) & 3
            if value == 0:
                return None  # Feature not supported by this device
            return CleanFluidStatus(value)
        return None

    @property
    def hatch_door_status(self) -> int | None:
        if self.dss:
            return (self.dss >> 12) & 7
        return None

    @property
    def dock_cool_fan_status(self) -> int | None:
        if self.dss:
            return (self.dss >> 15) & 3
        return None

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class StatusV2(RoborockBase):
    """
    This is a new version of the Status object.
    This is the result of GET_STATUS from the api.
    """

    msg_ver: int | None = None
    msg_seq: int | None = None
    state: RoborockStateCode | None = None
    battery: int | None = None
    clean_time: int | None = None
    clean_area: int | None = None
    error_code: RoborockErrorCode | None = None
    map_present: int | None = None
    in_cleaning: RoborockInCleaning | None = None
    in_returning: int | None = None
    in_fresh_state: int | None = None
    lab_status: int | None = None
    water_box_status: int | None = None
    back_type: int | None = None
    wash_phase: int | None = None
    wash_ready: int | None = None
    fan_power: int | None = None
    dnd_enabled: int | None = None
    map_status: int | None = None
    is_locating: int | None = None
    lock_status: int | None = None
    water_box_mode: int | None = None
    water_box_carriage_status: int | None = None
    mop_forbidden_enable: int | None = None
    camera_status: int | None = None
    is_exploring: int | None = None
    home_sec_status: int | None = None
    home_sec_enable_password: int | None = None
    adbumper_status: list[int] | None = None
    water_shortage_status: int | None = None
    dock_type: RoborockDockTypeCode | None = None
    dust_collection_status: int | None = None
    auto_dust_collection: int | None = None
    avoid_count: int | None = None
    mop_mode: int | None = None
    debug_mode: int | None = None
    collision_avoid_status: int | None = None
    switch_map_mode: int | None = None
    dock_error_status: RoborockDockErrorCode | None = None
    charge_status: int | None = None
    unsave_map_reason: int | None = None
    unsave_map_flag: int | None = None
    wash_status: int | None = None
    distance_off: int | None = None
    in_warmup: int | None = None
    dry_status: int | None = None
    rdt: int | None = None
    clean_percent: int | None = None
    rss: int | None = None
    dss: int | None = None
    common_status: int | None = None
    corner_clean_mode: int | None = None
    last_clean_t: int | None = None
    replenish_mode: int | None = None
    repeat: int | None = None
    kct: int | None = None
    subdivision_sets: int | None = None

    @property
    def square_meter_clean_area(self) -> float | None:
        return round(self.clean_area / 1000000, 1) if self.clean_area is not None else None

    @property
    def error_code_name(self) -> str | None:
        return self.error_code.name if self.error_code is not None else None

    @property
    def state_name(self) -> str | None:
        return self.state.name if self.state is not None else None

    @property
    def current_map(self) -> int | None:
        """Returns the current map ID if the map is present."""
        if self.map_status is not None:
            map_flag = self.map_status >> 2
            if map_flag != NO_MAP:
                return map_flag
        return None

    @property
    def clear_water_box_status(self) -> ClearWaterBoxStatus | None:
        if self.dss:
            return ClearWaterBoxStatus((self.dss >> 2) & 3)
        return None

    @property
    def dirty_water_box_status(self) -> DirtyWaterBoxStatus | None:
        if self.dss:
            return DirtyWaterBoxStatus((self.dss >> 4) & 3)
        return None

    @property
    def dust_bag_status(self) -> DustBagStatus | None:
        if self.dss:
            return DustBagStatus((self.dss >> 6) & 3)
        return None

    @property
    def water_box_filter_status(self) -> int | None:
        if self.dss:
            return (self.dss >> 8) & 3
        return None

    @property
    def clean_fluid_status(self) -> CleanFluidStatus | None:
        if self.dss:
            value = (self.dss >> 10) & 3
            if value == 0:
                return None  # Feature not supported by this device
            return CleanFluidStatus(value)
        return None

    @property
    def hatch_door_status(self) -> int | None:
        if self.dss:
            return (self.dss >> 12) & 7
        return None

    @property
    def dock_cool_fan_status(self) -> int | None:
        if self.dss:
            return (self.dss >> 15) & 3
        return None

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class S4MaxStatus(Status):
    fan_power: RoborockFanSpeedS6Pure | None = None
    water_box_mode: RoborockMopIntensityS7 | None = None
    mop_mode: RoborockMopModeS7 | None = None


@dataclass
class S5MaxStatus(Status):
    fan_power: RoborockFanSpeedS6Pure | None = None
    water_box_mode: RoborockMopIntensityS5Max | None = None


@dataclass
class Q7MaxStatus(Status):
    fan_power: RoborockFanSpeedQ7Max | None = None
    water_box_mode: RoborockMopIntensityQ7Max | None = None


@dataclass
class QRevoMasterStatus(Status):
    fan_power: RoborockFanSpeedQRevoMaster | None = None
    water_box_mode: RoborockMopIntensityQRevoMaster | None = None
    mop_mode: RoborockMopModeQRevoMaster | None = None


@dataclass
class QRevoCurvStatus(Status):
    fan_power: RoborockFanSpeedQRevoCurv | None = None
    water_box_mode: RoborockMopIntensityQRevoCurv | None = None
    mop_mode: RoborockMopModeQRevoCurv | None = None


@dataclass
class QRevoMaxVStatus(Status):
    fan_power: RoborockFanSpeedQRevoMaxV | None = None
    water_box_mode: RoborockMopIntensityQRevoMaxV | None = None
    mop_mode: RoborockMopModeQRevoMaxV | None = None


@dataclass
class S6MaxVStatus(Status):
    fan_power: RoborockFanSpeedS7MaxV | None = None
    water_box_mode: RoborockMopIntensityS6MaxV | None = None


@dataclass
class S6PureStatus(Status):
    fan_power: RoborockFanSpeedS6Pure | None = None


@dataclass
class S7MaxVStatus(Status):
    fan_power: RoborockFanSpeedS7MaxV | None = None
    water_box_mode: RoborockMopIntensityS7 | None = None
    mop_mode: RoborockMopModeS7 | None = None


@dataclass
class S7Status(Status):
    fan_power: RoborockFanSpeedS7 | None = None
    water_box_mode: RoborockMopIntensityS7 | None = None
    mop_mode: RoborockMopModeS7 | None = None


@dataclass
class S8ProUltraStatus(Status):
    fan_power: RoborockFanSpeedS7MaxV | None = None
    water_box_mode: RoborockMopIntensityS7 | None = None
    mop_mode: RoborockMopModeS8ProUltra | None = None


@dataclass
class S8Status(Status):
    fan_power: RoborockFanSpeedS7MaxV | None = None
    water_box_mode: RoborockMopIntensityS7 | None = None
    mop_mode: RoborockMopModeS8ProUltra | None = None


@dataclass
class P10Status(Status):
    fan_power: RoborockFanSpeedP10 | None = None
    water_box_mode: RoborockMopIntensityP10 | None = None
    mop_mode: RoborockMopModeS8ProUltra | None = None


@dataclass
class S8MaxvUltraStatus(Status):
    fan_power: RoborockFanSpeedS8MaxVUltra | None = None
    water_box_mode: RoborockMopIntensityS8MaxVUltra | None = None
    mop_mode: RoborockMopModeS8MaxVUltra | None = None


@dataclass
class Saros10RStatus(Status):
    fan_power: RoborockFanSpeedSaros10R | None = None
    water_box_mode: RoborockMopIntensitySaros10R | None = None
    mop_mode: RoborockMopModeSaros10R | None = None


@dataclass
class Saros10Status(Status):
    fan_power: RoborockFanSpeedSaros10 | None = None
    water_box_mode: RoborockMopIntensitySaros10 | None = None
    mop_mode: RoborockMopModeSaros10 | None = None


ModelStatus: dict[str, type[Status]] = {
    ROBOROCK_S4_MAX: S4MaxStatus,
    ROBOROCK_S5_MAX: S5MaxStatus,
    ROBOROCK_Q7_MAX: Q7MaxStatus,
    ROBOROCK_QREVO_MASTER: QRevoMasterStatus,
    ROBOROCK_QREVO_CURV: QRevoCurvStatus,
    ROBOROCK_S6: S6PureStatus,
    ROBOROCK_S6_MAXV: S6MaxVStatus,
    ROBOROCK_S6_PURE: S6PureStatus,
    ROBOROCK_S7_MAXV: S7MaxVStatus,
    ROBOROCK_S7: S7Status,
    ROBOROCK_S8: S8Status,
    ROBOROCK_S8_PRO_ULTRA: S8ProUltraStatus,
    ROBOROCK_G10S_PRO: S7MaxVStatus,
    ROBOROCK_G20S_Ultra: QRevoMasterStatus,
    ROBOROCK_P10: P10Status,
    # These likely are not correct,
    # but i am currently unable to do my typical reverse engineering/ get any data from users on this,
    # so this will be here in the mean time.
    ROBOROCK_QREVO_S: P10Status,
    ROBOROCK_QREVO_MAXV: QRevoMaxVStatus,
    ROBOROCK_QREVO_PRO: P10Status,
    ROBOROCK_S8_MAXV_ULTRA: S8MaxvUltraStatus,
    ROBOROCK_SAROS_10R: Saros10RStatus,
    ROBOROCK_SAROS_10: Saros10Status,
}


@dataclass
class DnDTimer(RoborockBaseTimer):
    """DnDTimer"""


@dataclass
class ValleyElectricityTimer(RoborockBaseTimer):
    """ValleyElectricityTimer"""


@dataclass
class CleanSummary(RoborockBase):
    clean_time: int | None = None
    clean_area: int | None = None
    clean_count: int | None = None
    dust_collection_count: int | None = None
    records: list[int] | None = None
    last_clean_t: int | None = None

    @property
    def square_meter_clean_area(self) -> float | None:
        """Returns the clean area in square meters."""
        if isinstance(self.clean_area, list | str):
            _LOGGER.warning(f"Clean area is a unexpected type! Please give the following in a issue: {self.clean_area}")
            return None
        return round(self.clean_area / 1000000, 1) if self.clean_area is not None else None

    def __repr__(self) -> str:
        """Return a string representation of the object including all attributes."""
        return _attr_repr(self)


@dataclass
class CleanRecord(RoborockBase):
    begin: int | None = None
    end: int | None = None
    duration: int | None = None
    area: int | None = None
    error: int | None = None
    complete: int | None = None
    start_type: RoborockStartType | None = None
    clean_type: RoborockCleanType | None = None
    finish_reason: RoborockFinishReason | None = None
    dust_collection_status: int | None = None
    avoid_count: int | None = None
    wash_count: int | None = None
    map_flag: int | None = None

    @property
    def square_meter_area(self) -> float | None:
        return round(self.area / 1000000, 1) if self.area is not None else None

    @property
    def begin_datetime(self) -> datetime.datetime | None:
        return datetime.datetime.fromtimestamp(self.begin).astimezone(datetime.UTC) if self.begin else None

    @property
    def end_datetime(self) -> datetime.datetime | None:
        return datetime.datetime.fromtimestamp(self.end).astimezone(datetime.UTC) if self.end else None

    def __repr__(self) -> str:
        return _attr_repr(self)


class CleanSummaryWithDetail(CleanSummary):
    """CleanSummary with the last CleanRecord included."""

    last_clean_record: CleanRecord | None = None


@dataclass
class Consumable(RoborockBase):
    main_brush_work_time: int | None = None
    side_brush_work_time: int | None = None
    filter_work_time: int | None = None
    filter_element_work_time: int | None = None
    sensor_dirty_time: int | None = None
    strainer_work_times: int | None = None
    dust_collection_work_times: int | None = None
    cleaning_brush_work_times: int | None = None
    moproller_work_time: int | None = None

    @property
    def main_brush_time_left(self) -> int | None:
        return MAIN_BRUSH_REPLACE_TIME - self.main_brush_work_time if self.main_brush_work_time is not None else None

    @property
    def side_brush_time_left(self) -> int | None:
        return SIDE_BRUSH_REPLACE_TIME - self.side_brush_work_time if self.side_brush_work_time is not None else None

    @property
    def filter_time_left(self) -> int | None:
        return FILTER_REPLACE_TIME - self.filter_work_time if self.filter_work_time is not None else None

    @property
    def sensor_time_left(self) -> int | None:
        return SENSOR_DIRTY_REPLACE_TIME - self.sensor_dirty_time if self.sensor_dirty_time is not None else None

    @property
    def strainer_time_left(self) -> int | None:
        return STRAINER_REPLACE_TIME - self.strainer_work_times if self.strainer_work_times is not None else None

    @property
    def dust_collection_time_left(self) -> int | None:
        return (
            DUST_COLLECTION_REPLACE_TIME - self.dust_collection_work_times
            if self.dust_collection_work_times is not None
            else None
        )

    @property
    def cleaning_brush_time_left(self) -> int | None:
        return (
            CLEANING_BRUSH_REPLACE_TIME - self.cleaning_brush_work_times
            if self.cleaning_brush_work_times is not None
            else None
        )

    @property
    def mop_roller_time_left(self) -> int | None:
        return MOP_ROLLER_REPLACE_TIME - self.moproller_work_time if self.moproller_work_time is not None else None

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class MultiMapsListRoom(RoborockBase):
    id: int | None = None
    tag: int | None = None
    iot_name_id: str | None = None
    iot_name: str | None = None

    @property
    def named_room_mapping(self) -> NamedRoomMapping | None:
        """Returns a NamedRoomMapping object if valid."""
        if self.id is None or self.iot_name_id is None:
            return None
        return NamedRoomMapping(
            segment_id=self.id,
            iot_id=self.iot_name_id,
            raw_name=self.iot_name,
        )


@dataclass
class MultiMapsListMapInfoBakMaps(RoborockBase):
    mapflag: Any | None = None
    add_time: Any | None = None


@dataclass
class MultiMapsListMapInfo(RoborockBase):
    map_flag: int
    name: str
    add_time: Any | None = None
    length: Any | None = None
    bak_maps: list[MultiMapsListMapInfoBakMaps] | None = None
    rooms: list[MultiMapsListRoom] | None = None

    @property
    def mapFlag(self) -> int:
        """Alias for map_flag, returns the map flag as an integer."""
        return self.map_flag

    @property
    def rooms_map(self) -> dict[int, NamedRoomMapping]:
        """Returns a dictionary of room mappings by segment id."""
        return {
            room.id: mapping
            for room in self.rooms or ()
            if room.id is not None and (mapping := room.named_room_mapping) is not None
        }


@dataclass
class MultiMapsList(RoborockBase):
    max_multi_map: int | None = None
    max_bak_map: int | None = None
    multi_map_count: int | None = None
    map_info: list[MultiMapsListMapInfo] | None = None


@dataclass
class SmartWashParams(RoborockBase):
    smart_wash: int | None = None
    wash_interval: int | None = None


@dataclass
class DustCollectionMode(RoborockBase):
    mode: RoborockDockDustCollectionModeCode | None = None


@dataclass
class WashTowelMode(RoborockBase):
    wash_mode: RoborockDockWashTowelModeCode | None = None


@dataclass
class NetworkInfo(RoborockBase):
    ip: str
    ssid: str | None = None
    mac: str | None = None
    bssid: str | None = None
    rssi: int | None = None


@dataclass
class AppInitStatusLocalInfo(RoborockBase):
    location: str
    bom: str | None = None
    featureset: int | None = None
    language: str | None = None
    logserver: str | None = None
    wifiplan: str | None = None
    timezone: str | None = None
    name: str | None = None


@dataclass
class AppInitStatus(RoborockBase):
    local_info: AppInitStatusLocalInfo
    feature_info: list[int]
    new_feature_info: int
    new_feature_info_str: str = ""
    new_feature_info_2: int | None = None
    carriage_type: int | None = None
    dsp_version: str | None = None


@dataclass
class ChildLockStatus(RoborockBase):
    lock_status: int = 0


@dataclass
class FlowLedStatus(RoborockBase):
    status: int = 0


@dataclass
class LedStatus(RoborockBase):
    status: int = 0



"""This module implements a file-backed cache for device information.

This module provides a `FileCache` class that implements the `Cache` protocol
to store and retrieve cached device information from a file on disk. This allows
persistent caching of device data across application restarts.
"""

import asyncio
import pathlib
import pickle
from collections.abc import Callable
from typing import Any

from .cache import Cache, CacheData


class FileCache(Cache):
    """File backed cache implementation."""

    def __init__(
        self,
        file_path: pathlib.Path,
        init_fn: Callable[[], CacheData] = CacheData,
        serialize_fn: Callable[[Any], bytes] = pickle.dumps,
        deserialize_fn: Callable[[bytes], Any] = pickle.loads,
    ) -> None:
        """Initialize the file cache with the given file path."""
        self._init_fn = init_fn
        self._file_path = file_path
        self._cache_data: CacheData | None = None
        self._serialize_fn = serialize_fn
        self._deserialize_fn = deserialize_fn

    async def get(self) -> CacheData:
        """Get cached value."""
        if self._cache_data is not None:
            return self._cache_data
        data = await load_value(self._file_path, self._deserialize_fn)
        if data is not None and not isinstance(data, CacheData):
            raise TypeError(f"Invalid cache data loaded from {self._file_path}")

        self._cache_data = data or self._init_fn()
        return self._cache_data

    async def set(self, value: CacheData) -> None:  # type: ignore[override]
        """Set value in the cache."""
        self._cache_data = value

    async def flush(self) -> None:
        """Flush the cache to disk."""
        if self._cache_data is None:
            return
        await store_value(self._file_path, self._cache_data, self._serialize_fn)


async def store_value(file_path: pathlib.Path, value: Any, serialize_fn: Callable[[Any], bytes] = pickle.dumps) -> None:
    """Store a value to the given file path."""

    def _store_to_disk(file_path: pathlib.Path, value: Any) -> None:
        with open(file_path, "wb") as f:
            data = serialize_fn(value)
            f.write(data)

    await asyncio.to_thread(_store_to_disk, file_path, value)


async def load_value(file_path: pathlib.Path, deserialize_fn: Callable[[bytes], Any] = pickle.loads) -> Any | None:
    """Load a value from the given file path."""

    def _load_from_disk(file_path: pathlib.Path) -> Any | None:
        if not file_path.exists():
            return None
        with open(file_path, "rb") as f:
            data = f.read()
            if not data:  # 空文件视为不存在
                return None
            return deserialize_fn(data)

    return await asyncio.to_thread(_load_from_disk, file_path)



"""Thin wrapper around the MQTT channel for Roborock B01 Q10 devices."""

from __future__ import annotations

import logging
from collections.abc import AsyncGenerator
from typing import Any

from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP
from roborock.devices.transport.mqtt_channel import MqttChannel
from roborock.exceptions import RoborockException
from roborock.protocols.b01_q10_protocol import (
    ParamsType,
    decode_rpc_response,
    encode_mqtt_payload,
)

_LOGGER = logging.getLogger(__name__)


async def stream_decoded_responses(
    mqtt_channel: MqttChannel,
) -> AsyncGenerator[dict[B01_Q10_DP, Any], None]:
    """Stream decoded DPS messages received via MQTT."""

    async for response_message in mqtt_channel.subscribe_stream():
        try:
            decoded_dps = decode_rpc_response(response_message)
        except RoborockException as ex:
            _LOGGER.debug(
                "Failed to decode B01 Q10 RPC response: %s: %s",
                response_message,
                ex,
            )
            continue
        yield decoded_dps


async def send_command(
    mqtt_channel: MqttChannel,
    command: B01_Q10_DP,
    params: ParamsType,
) -> None:
    """Send a command on the MQTT channel, without waiting for a response"""
    _LOGGER.debug("Sending B01 MQTT command: cmd=%s params=%s", command, params)
    roborock_message = encode_mqtt_payload(command, params)
    _LOGGER.debug("Sending MQTT message: %s", roborock_message)
    try:
        await mqtt_channel.publish(roborock_message)
    except RoborockException as ex:
        _LOGGER.debug(
            "Error sending B01 decoded command (method=%s params=%s): %s",
            command,
            params,
            ex,
        )
        raise



"""Routines trait for V1 devices."""

from roborock.data.containers import HomeDataScene
from roborock.web_api import UserWebApiClient


class RoutinesTrait:
    """Trait for interacting with routines."""

    def __init__(self, device_id: str, web_api: UserWebApiClient) -> None:
        """Initialize the routines trait."""
        self._device_id = device_id
        self._web_api = web_api

    async def get_routines(self) -> list[HomeDataScene]:
        """Get available routines."""
        return await self._web_api.get_routines(self._device_id)

    async def execute_routine(self, routine_id: int) -> None:
        """Execute a routine by its ID.

        Technically, routines are per-device, but the API does not
        require the device ID to execute them. This can execute a
        routine for any device but it is exposed here for convenience.
        """
        await self._web_api.execute_routine(routine_id)



"""Trait for managing consumable attributes.

A consumable attribute is one that is expected to be replaced or refilled
periodically, such as filters, brushes, etc.
"""

from enum import StrEnum
from typing import Self

from roborock.data import Consumable
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand

__all__ = [
    "ConsumableTrait",
]


class ConsumableAttribute(StrEnum):
    """Enum for consumable attributes."""

    SENSOR_DIRTY_TIME = "sensor_dirty_time"
    FILTER_WORK_TIME = "filter_work_time"
    SIDE_BRUSH_WORK_TIME = "side_brush_work_time"
    MAIN_BRUSH_WORK_TIME = "main_brush_work_time"

    @classmethod
    def from_str(cls, value: str) -> Self:
        """Create a ConsumableAttribute from a string value."""
        for member in cls:
            if member.value == value:
                return member
        raise ValueError(f"Unknown ConsumableAttribute: {value}")


class ConsumableTrait(Consumable, common.V1TraitMixin):
    """Trait for managing consumable attributes on Roborock devices.

    After the first refresh, you can tell what consumables are supported by
    checking which attributes are not None.
    """

    command = RoborockCommand.GET_CONSUMABLE

    async def reset_consumable(self, consumable: ConsumableAttribute) -> None:
        """Reset a specific consumable attribute on the device."""
        await self.rpc_channel.send_command(RoborockCommand.RESET_CONSUMABLE, params=[consumable.value])
        await self.refresh()



from roborock.data import ValleyElectricityTimer
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand

_ENABLED_PARAM = "enabled"


class ValleyElectricityTimerTrait(ValleyElectricityTimer, common.V1TraitMixin, common.RoborockSwitchBase):
    """Trait for managing Valley Electricity Timer settings on Roborock devices."""

    command = RoborockCommand.GET_VALLEY_ELECTRICITY_TIMER
    requires_feature = "is_supported_valley_electricity"

    @property
    def is_on(self) -> bool:
        """Return whether the Valley Electricity Timer is enabled."""
        return self.enabled == 1

    async def set_timer(self, timer: ValleyElectricityTimer) -> None:
        """Set the Valley Electricity Timer settings of the device."""
        await self.rpc_channel.send_command(RoborockCommand.SET_VALLEY_ELECTRICITY_TIMER, params=timer.as_list())
        await self.refresh()

    async def clear_timer(self) -> None:
        """Clear the Valley Electricity Timer settings of the device."""
        await self.rpc_channel.send_command(RoborockCommand.CLOSE_VALLEY_ELECTRICITY_TIMER)
        await self.refresh()

    async def enable(self) -> None:
        """Enable the Valley Electricity Timer settings of the device."""
        await self.rpc_channel.send_command(
            RoborockCommand.SET_VALLEY_ELECTRICITY_TIMER,
            params=self.as_list(),
        )
        # Optimistic update to avoid an extra refresh
        self.enabled = 1

    async def disable(self) -> None:
        """Disable the Valley Electricity Timer settings of the device."""
        await self.rpc_channel.send_command(
            RoborockCommand.CLOSE_VALLEY_ELECTRICITY_TIMER,
        )
        # Optimistic update to avoid an extra refresh
        self.enabled = 0



"""
连接管理模块

负责：
- 从 JSON 缓存加载 UserData
- 创建 DeviceManager
- 设备发现和模糊匹配
- 连接生命周期管理
"""

import json
import os
from contextlib import asynccontextmanager
from pathlib import Path
from typing import Any

from roborock.data import UserData
from roborock.devices.device_manager import create_device_manager, UserParams
from roborock.devices.file_cache import FileCache


class AuthExpiredError(Exception):
    """认证已过期，需要重新运行 setup_auth"""

    def __init__(self):
        super().__init__("认证已过期，请重新运行 `roborock-mcp auth`")


class NoDeviceError(Exception):
    """未发现设备"""

    def __init__(self):
        super().__init__("未发现任何设备，请确认设备已绑定到账户")


class AmbiguousDeviceError(Exception):
    """设备名称匹配到多个设备"""

    def __init__(self, name: str, matches: list[str]):
        self.name = name
        self.matches = matches
        super().__init__(
            f"设备名称 '{name}' 匹配到多个设备: {matches}，请使用更精确的名称"
        )


class DeviceNotFoundError(Exception):
    """设备名称未匹配到任何设备"""

    def __init__(self, name: str, available: list[str]):
        self.name = name
        self.available = available
        super().__init__(
            f"未找到设备 '{name}'，可用设备: {available}"
        )


class ConnectionManager:
    """连接管理器"""

    def __init__(self, auth_cache_path: Path, device_cache_path: Path):
        self._auth_cache_path = auth_cache_path
        self._device_cache_path = device_cache_path
        self._manager: Any = None  # DeviceManager
        self._cache: FileCache | None = None
        self._user_params: UserParams | None = None

    @classmethod
    def from_env(cls) -> "ConnectionManager":
        """从环境变量创建连接管理器"""
        auth_cache = Path(
            os.environ.get(
                "ROBOROCK_AUTH_CACHE",
                Path.home() / ".cache" / "roborock-mcp" / "user_data.json"
            )
        )
        device_cache = Path(
            os.environ.get(
                "ROBOROCK_DEVICE_CACHE",
                Path.home() / ".cache" / "roborock-mcp" / "device_cache.pkl"
            )
        )
        return cls(auth_cache, device_cache)

    async def ensure_connected(self) -> None:
        """确保已连接，如果未连接则建立连接"""
        if self._manager is not None:
            return

        # 1. 从 JSON 文件加载 UserData
        if not self._auth_cache_path.exists():
            raise AuthExpiredError()

        try:
            with open(self._auth_cache_path) as f:
                data = json.load(f)
            email = data["email"]
            user_data = UserData.from_dict(data["user_data"])
        except (json.JSONDecodeError, KeyError) as e:
            raise AuthExpiredError() from e

        # 2. 构建 UserParams（不传 base_url，让 SDK 自动发现）
        self._user_params = UserParams(
            username=email,
            user_data=user_data,
        )

        # 3. 创建缓存目录
        self._device_cache_path.parent.mkdir(parents=True, exist_ok=True)

        # 4. 创建 FileCache
        self._cache = FileCache(self._device_cache_path)

        # 5. 创建 DeviceManager（内部已调用 discover_devices）
        self._manager = await create_device_manager(
            self._user_params,
            cache=self._cache
        )

        # 6. 检查设备列表
        devices = await self._manager.get_devices()
        if not devices:
            raise NoDeviceError()

    async def list_devices(self) -> list[Any]:
        """返回所有已发现设备"""
        if self._manager is None:
            await self.ensure_connected()
        return await self._manager.get_devices()

    async def get_device(self, device_name: str | None = None) -> Any:
        """
        获取指定设备

        Args:
            device_name: 设备名称（模糊匹配），为 None 时自动选择单设备

        Returns:
            RoborockDevice

        Raises:
            NoDeviceError: 无设备
            AmbiguousDeviceError: 名称匹配到多个设备
            DeviceNotFoundError: 名称未匹配到任何设备
        """
        devices = await self.list_devices()

        # 单设备且未指定名称时，自动选择
        if device_name is None:
            if len(devices) == 1:
                return devices[0]
            else:
                available = [d.name for d in devices]
                raise ValueError(
                    f"发现 {len(devices)} 个设备，请指定 device_name。可用设备: {available}"
                )

        # 模糊匹配
        matches = []
        for device in devices:
            if device_name.lower() in device.name.lower():
                matches.append(device)

        if len(matches) == 0:
            available = [d.name for d in devices]
            raise DeviceNotFoundError(device_name, available)

        if len(matches) > 1:
            raise AmbiguousDeviceError(
                device_name,
                [d.name for d in matches]
            )

        return matches[0]

    async def close(self) -> None:
        """关闭连接，先 flush 缓存再关闭"""
        if self._cache is not None:
            await self._cache.flush()
        if self._manager is not None:
            await self._manager.close()


@asynccontextmanager
async def lifespan(server: Any):
    """MCP server lifespan 管理"""
    conn = ConnectionManager.from_env()
    await conn.ensure_connected()
    yield {"conn": conn}
    await conn.close()



"""V1 Channel for Roborock devices.

This module provides a unified channel interface for V1 protocol devices,
handling both MQTT and local connections with automatic fallback.
"""

import asyncio
import datetime
import logging
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, TypeVar

from roborock.data import HomeDataDevice, NetworkInfo, RoborockBase, UserData
from roborock.devices.cache import DeviceCache
from roborock.devices.transport.channel import Channel
from roborock.devices.transport.local_channel import LocalChannel, LocalSession, create_local_session
from roborock.devices.transport.mqtt_channel import MqttChannel
from roborock.exceptions import RoborockException
from roborock.mqtt.health_manager import HealthManager
from roborock.mqtt.session import MqttParams, MqttSession
from roborock.protocols.v1_protocol import (
    CommandType,
    MapResponse,
    ParamsType,
    RequestMessage,
    ResponseData,
    ResponseMessage,
    SecurityData,
    V1RpcChannel,
    create_map_response_decoder,
    create_security_data,
    decode_rpc_response,
)
from roborock.roborock_message import RoborockMessage, RoborockMessageProtocol
from roborock.roborock_typing import RoborockCommand
from roborock.util import RoborockLoggerAdapter

_LOGGER = logging.getLogger(__name__)

__all__ = [
    "create_v1_channel",
]

_T = TypeVar("_T", bound=RoborockBase)
_TIMEOUT = 10.0


# Exponential backoff parameters for reconnecting to local
MIN_RECONNECT_INTERVAL = datetime.timedelta(minutes=1)
MAX_RECONNECT_INTERVAL = datetime.timedelta(minutes=10)
RECONNECT_MULTIPLIER = 1.5
# After this many hours, the network info is refreshed
NETWORK_INFO_REFRESH_INTERVAL = datetime.timedelta(hours=12)
# Interval to check that the local connection is healthy
LOCAL_CONNECTION_CHECK_INTERVAL = datetime.timedelta(seconds=15)


@dataclass(frozen=True)
class RpcStrategy:
    """Strategy for encoding/sending/decoding RPC commands."""

    name: str  # For debug logging
    channel: LocalChannel | MqttChannel
    encoder: Callable[[RequestMessage], RoborockMessage]
    decoder: Callable[[RoborockMessage], ResponseMessage | MapResponse | None]
    health_manager: HealthManager | None = None


class RpcChannel(V1RpcChannel):
    """Provides an RPC interface around a pub/sub transport channel."""

    def __init__(self, rpc_strategies_cb: Callable[[], list[RpcStrategy]], logger: RoborockLoggerAdapter) -> None:
        """Initialize the RpcChannel with an ordered list of strategies."""
        self._rpc_strategies_cb = rpc_strategies_cb
        self._logger = logger

    async def send_command(
        self,
        method: CommandType,
        *,
        response_type: type[_T] | None = None,
        params: ParamsType = None,
    ) -> _T | Any:
        """Send a command and return either a decoded or parsed response."""
        request = RequestMessage(method, params=params)

        # Try each channel in order until one succeeds
        last_exception = None
        for strategy in self._rpc_strategies_cb():
            try:
                decoded_response = await self._send_rpc(strategy, request, self._logger)
            except RoborockException as e:
                self._logger.debug("Command %s failed on %s channel: %s", method, strategy.name, e)
                last_exception = e
            except Exception as e:
                self._logger.exception("Unexpected error sending command %s on %s channel", method, strategy.name)
                last_exception = RoborockException(f"Unexpected error: {e}")
            else:
                if response_type is not None:
                    if not isinstance(decoded_response, dict):
                        raise RoborockException(
                            f"Expected dict response to parse {response_type.__name__}, got {type(decoded_response)}"
                        )
                    return response_type.from_dict(decoded_response)
                return decoded_response

        raise last_exception or RoborockException("No available connection to send command")

    @staticmethod
    async def _send_rpc(
        strategy: RpcStrategy, request: RequestMessage, logger: RoborockLoggerAdapter
    ) -> ResponseData | bytes:
        """Send a command and return a decoded response type.

        This provides an RPC interface over a given channel strategy. The device
        channel only supports publish and subscribe, so this function handles
        associating requests with their corresponding responses.
        """
        future: asyncio.Future[ResponseData | bytes] = asyncio.Future()
        logger.debug(
            "Sending command (%s, request_id=%s): %s, params=%s",
            strategy.name,
            request.request_id,
            request.method,
            request.params,
        )

        message = strategy.encoder(request)

        def find_response(response_message: RoborockMessage) -> None:
            try:
                decoded = strategy.decoder(response_message)
            except RoborockException as ex:
                logger.debug("Exception while decoding message (%s): %s", response_message, ex)
                return
            if decoded is None:
                return
            logger.debug("Received response (%s, request_id=%s)", strategy.name, decoded.request_id)
            if decoded.request_id == request.request_id:
                if isinstance(decoded, ResponseMessage) and decoded.api_error:
                    future.set_exception(decoded.api_error)
                else:
                    future.set_result(decoded.data)

        unsub = await strategy.channel.subscribe(find_response)
        try:
            await strategy.channel.publish(message)
            result = await asyncio.wait_for(future, timeout=_TIMEOUT)
        except TimeoutError as ex:
            if strategy.health_manager:
                await strategy.health_manager.on_timeout()
            future.cancel()
            raise RoborockException(f"Command timed out after {_TIMEOUT}s") from ex
        finally:
            unsub()
        if strategy.health_manager:
            await strategy.health_manager.on_success()
        return result


class V1Channel(Channel):
    """Unified V1 protocol channel with automatic MQTT/local connection handling.

    This channel abstracts away the complexity of choosing between MQTT and local
    connections, and provides high-level V1 protocol methods. It automatically
    handles connection setup, fallback logic, and protocol encoding/decoding.
    """

    def __init__(
        self,
        device_uid: str,
        security_data: SecurityData,
        mqtt_channel: MqttChannel,
        local_session: LocalSession,
        device_cache: DeviceCache,
    ) -> None:
        """Initialize the V1Channel."""
        self._device_uid = device_uid
        self._logger = RoborockLoggerAdapter(duid=device_uid, logger=_LOGGER)
        self._security_data = security_data
        self._mqtt_channel = mqtt_channel
        self._local_session = local_session
        self._local_channel: LocalChannel | None = None
        self._mqtt_unsub: Callable[[], None] | None = None
        self._local_unsub: Callable[[], None] | None = None
        self._callback: Callable[[RoborockMessage], None] | None = None
        self._device_cache = device_cache
        self._reconnect_task: asyncio.Task[None] | None = None
        self._last_network_info_refresh: datetime.datetime | None = None

    @property
    def is_connected(self) -> bool:
        """Return whether any connection is available."""
        return self.is_mqtt_connected or self.is_local_connected

    @property
    def is_local_connected(self) -> bool:
        """Return whether local connection is available."""
        return self._local_channel is not None and self._local_channel.is_connected

    @property
    def is_mqtt_connected(self) -> bool:
        """Return whether MQTT connection is available."""
        return self._mqtt_channel.is_connected

    @property
    def rpc_channel(self) -> V1RpcChannel:
        """Return the combined RPC channel that prefers local with a fallback to MQTT.

        The returned V1RpcChannel may be long lived and will respect the
        current connection state of the underlying channels.
        """

        def rpc_strategies_cb() -> list[RpcStrategy]:
            strategies = []
            if local_rpc_strategy := self._create_local_rpc_strategy():
                strategies.append(local_rpc_strategy)
            strategies.append(self._create_mqtt_rpc_strategy())
            return strategies

        return RpcChannel(rpc_strategies_cb, self._logger)

    @property
    def mqtt_rpc_channel(self) -> V1RpcChannel:
        """Return the MQTT-only RPC channel.

        The returned V1RpcChannel may be long lived and will respect the
        current connection state of the underlying channels.
        """
        return RpcChannel(lambda: [self._create_mqtt_rpc_strategy()], self._logger)

    @property
    def map_rpc_channel(self) -> V1RpcChannel:
        """Return the map RPC channel used for fetching map content."""
        decoder = create_map_response_decoder(security_data=self._security_data)
        return RpcChannel(lambda: [self._create_mqtt_rpc_strategy(decoder)], self._logger)

    def _create_local_rpc_strategy(self) -> RpcStrategy | None:
        """Create the RPC strategy for local transport."""
        if self._local_channel is None or not self.is_local_connected:
            return None
        return RpcStrategy(
            name="local",
            channel=self._local_channel,
            encoder=self._local_encoder,
            decoder=decode_rpc_response,
        )

    def _local_encoder(self, x: RequestMessage) -> RoborockMessage:
        """Encode a request message for local transport.

        This will read the current local channel's protocol version which
        changes as the protocol version is discovered.
        """
        if self._local_channel is None:
            raise ValueError("Local channel unavailable for encoding")
        return x.encode_message(
            RoborockMessageProtocol.GENERAL_REQUEST,
            version=self._local_channel.protocol_version,
        )

    def _create_mqtt_rpc_strategy(self, decoder: Callable[[RoborockMessage], Any] = decode_rpc_response) -> RpcStrategy:
        """Create the RPC strategy for MQTT transport with optional custom decoder."""
        return RpcStrategy(
            name="mqtt",
            channel=self._mqtt_channel,
            encoder=lambda x: x.encode_message(
                RoborockMessageProtocol.RPC_REQUEST,
                security_data=self._security_data,
            ),
            decoder=decoder,
            health_manager=self._mqtt_channel.health_manager,
        )

    async def subscribe(self, callback: Callable[[RoborockMessage], None]) -> Callable[[], None]:
        """Subscribe to all messages from the device.

        This will first attempt to establish a local connection to the device
        using cached network information if available. If that fails, it will
        fall back to using the MQTT connection.

        A background task will be started to monitor and maintain the local
        connection, attempting to reconnect as needed.

        Args:
            callback: Callback to invoke for each received message.

        Returns:
            Unsubscribe function to stop receiving messages and clean up resources.
        """
        if self._callback is not None:
            raise ValueError("Only one subscription allowed at a time")

        # Make an initial, optimistic attempt to connect to local with the
        # cache. The cache information will be refreshed by the background task.
        try:
            await self._local_connect(prefer_cache=True)
        except RoborockException as err:
            self._logger.debug("First local connection attempt failed, will retry: %s", err)

        # Start a background task to manage the local connection health. This
        # happens independent of whether we were able to connect locally now.
        if self._reconnect_task is None:
            loop = asyncio.get_running_loop()
            self._reconnect_task = loop.create_task(self._background_reconnect())

        if not self.is_local_connected:
            # We were not able to connect locally, so fallback to MQTT and at least
            # establish that connection explicitly. If this fails then raise an
            # error and let the caller know we failed to subscribe.
            self._mqtt_unsub = await self._mqtt_channel.subscribe(self._on_mqtt_message)
            self._logger.debug("V1Channel connected to device via MQTT")

        def unsub() -> None:
            """Unsubscribe from all messages."""
            if self._reconnect_task:
                self._reconnect_task.cancel()
                self._reconnect_task = None
            if self._mqtt_unsub:
                self._mqtt_unsub()
                self._mqtt_unsub = None
            if self._local_unsub:
                self._local_unsub()
                self._local_unsub = None
            self._callback = None  # 修复：清除 callback，允许重新订阅
            self._logger.debug("Unsubscribed from device")

        self._callback = callback
        return unsub

    async def _get_networking_info(self, *, prefer_cache: bool = True) -> NetworkInfo:
        """Retrieve networking information for the device.

        This is a cloud only command used to get the local device's IP address.
        """
        device_cache_data = await self._device_cache.get()

        if prefer_cache and device_cache_data.network_info:
            self._logger.debug("Using cached network info")
            return device_cache_data.network_info
        try:
            network_info = await self.mqtt_rpc_channel.send_command(
                RoborockCommand.GET_NETWORK_INFO, response_type=NetworkInfo
            )
        except RoborockException as e:
            self._logger.debug("Error fetching network info for device")
            if device_cache_data.network_info:
                self._logger.debug("Falling back to cached network info after error")
                return device_cache_data.network_info
            raise RoborockException(f"Network info failed for device {self._device_uid}") from e
        self._logger.debug("Network info for device: %s", network_info)
        self._last_network_info_refresh = datetime.datetime.now(datetime.UTC)

        device_cache_data = await self._device_cache.get()
        device_cache_data.network_info = network_info
        await self._device_cache.set(device_cache_data)
        return network_info

    async def _local_connect(self, *, prefer_cache: bool = True) -> None:
        """Set up local connection if possible."""
        self._logger.debug("Attempting to connect to local channel (prefer_cache=%s)", prefer_cache)
        networking_info = await self._get_networking_info(prefer_cache=prefer_cache)
        host = networking_info.ip
        self._logger.debug("Connecting to local channel at %s", host)
        # Create a new local channel and connect
        local_channel = self._local_session(host)
        try:
            await local_channel.connect()
        except RoborockException as e:
            raise RoborockException(f"Error connecting to local device {self._device_uid}: {e}") from e
        # Wire up the new channel
        self._local_channel = local_channel
        self._local_unsub = await self._local_channel.subscribe(self._on_local_message)
        self._logger.info("Connected to local channel successfully")

    async def _background_reconnect(self) -> None:
        """Task to run in the background to manage the local connection."""
        self._logger.debug("Starting background task to manage local connection")
        reconnect_backoff = MIN_RECONNECT_INTERVAL
        local_connect_failures = 0

        while True:
            try:
                if self.is_local_connected:
                    await asyncio.sleep(LOCAL_CONNECTION_CHECK_INTERVAL.total_seconds())
                    continue

                # Not connected, so wait with backoff before trying to connect.
                # The first time through, we don't sleep, we just try to connect.
                local_connect_failures += 1
                if local_connect_failures > 1:
                    await asyncio.sleep(reconnect_backoff.total_seconds())
                    reconnect_backoff = min(reconnect_backoff * RECONNECT_MULTIPLIER, MAX_RECONNECT_INTERVAL)

                use_cache = self._should_use_cache(local_connect_failures)
                await self._local_connect(prefer_cache=use_cache)
                # Reset backoff and failures on success
                reconnect_backoff = MIN_RECONNECT_INTERVAL
                local_connect_failures = 0

            except asyncio.CancelledError:
                self._logger.debug("Background reconnect task cancelled")
                if self._local_channel:
                    self._local_channel.close()
                return
            except RoborockException as err:
                self._logger.debug("Background reconnect failed: %s", err)
            except Exception:
                self._logger.exception("Unhandled exception in background reconnect task")

    def _should_use_cache(self, local_connect_failures: int) -> bool:
        """Determine whether to use cached network info on retries.

        On the first retry we'll avoid the cache to handle the case where
        the network ip may have recently changed. Otherwise, use the cache
        if available then expire at some point.
        """
        if local_connect_failures == 1:
            return False
        elif self._last_network_info_refresh and (
            datetime.datetime.now(datetime.UTC) - self._last_network_info_refresh > NETWORK_INFO_REFRESH_INTERVAL
        ):
            return False
        return True

    def _on_mqtt_message(self, message: RoborockMessage) -> None:
        """Handle incoming MQTT messages."""
        self._logger.debug("V1Channel received MQTT message: %s", message)
        if self._callback:
            self._callback(message)

    def _on_local_message(self, message: RoborockMessage) -> None:
        """Handle incoming local messages."""
        self._logger.debug("V1Channel received local message: %s", message)
        if self._callback:
            self._callback(message)


def create_v1_channel(
    user_data: UserData,
    mqtt_params: MqttParams,
    mqtt_session: MqttSession,
    device: HomeDataDevice,
    device_cache: DeviceCache,
) -> V1Channel:
    """Create a V1Channel for the given device."""
    security_data = create_security_data(user_data.rriot)
    mqtt_channel = MqttChannel(mqtt_session, device.duid, device.local_key, user_data.rriot, mqtt_params)
    local_session = create_local_session(device.local_key, device.duid)
    return V1Channel(
        device.duid,
        security_data,
        mqtt_channel,
        local_session=local_session,
        device_cache=device_cache,
    )



"""Traits for Q7 B01 devices.
Potentially other devices may fall into this category in the future."""

from typing import Any

from roborock import B01Props
from roborock.data.b01_q7.b01_q7_code_mappings import (
    CleanPathPreferenceMapping,
    CleanRepeatMapping,
    CleanTaskTypeMapping,
    CleanTypeMapping,
    SCDeviceCleanParam,
    SCWindMapping,
    WaterLevelMapping,
)
from roborock.devices.rpc.b01_q7_channel import send_decoded_command
from roborock.devices.traits import Trait
from roborock.devices.transport.mqtt_channel import MqttChannel
from roborock.protocols.b01_q7_protocol import CommandType, ParamsType, Q7RequestMessage
from roborock.roborock_message import RoborockB01Props
from roborock.roborock_typing import RoborockB01Q7Methods

from .clean_summary import CleanSummaryTrait

__all__ = [
    "Q7PropertiesApi",
    "CleanSummaryTrait",
]


class Q7PropertiesApi(Trait):
    """API for interacting with B01 devices."""

    clean_summary: CleanSummaryTrait
    """Trait for clean records / clean summary (Q7 `service.get_record_list`)."""

    def __init__(self, channel: MqttChannel) -> None:
        """Initialize the B01Props API."""
        self._channel = channel
        self.clean_summary = CleanSummaryTrait(channel)

    async def query_values(self, props: list[RoborockB01Props]) -> B01Props | None:
        """Query the device for the values of the given Q7 properties."""
        result = await self.send(
            RoborockB01Q7Methods.GET_PROP,
            {"property": props},
        )
        if not isinstance(result, dict):
            raise TypeError(f"Unexpected response type for GET_PROP: {type(result).__name__}: {result!r}")
        return B01Props.from_dict(result)

    async def set_prop(self, prop: RoborockB01Props, value: Any) -> None:
        """Set a property on the device."""
        await self.send(
            command=RoborockB01Q7Methods.SET_PROP,
            params={prop: value},
        )

    async def set_fan_speed(self, fan_speed: SCWindMapping) -> None:
        """Set the fan speed (wind)."""
        await self.set_prop(RoborockB01Props.WIND, fan_speed.code)

    async def set_water_level(self, water_level: WaterLevelMapping) -> None:
        """Set the water level (water)."""
        await self.set_prop(RoborockB01Props.WATER, water_level.code)

    async def set_mode(self, mode: CleanTypeMapping) -> None:
        """Set the cleaning mode (vacuum, mop, or vacuum and mop)."""
        await self.set_prop(RoborockB01Props.MODE, mode.code)

    async def set_clean_path_preference(self, preference: CleanPathPreferenceMapping) -> None:
        """Set the cleaning path preference (route)."""
        await self.set_prop(RoborockB01Props.CLEAN_PATH_PREFERENCE, preference.code)

    async def set_repeat_state(self, repeat: CleanRepeatMapping) -> None:
        """Set the cleaning repeat state (cycles)."""
        await self.set_prop(RoborockB01Props.REPEAT_STATE, repeat.code)

    async def start_clean(self) -> None:
        """Start cleaning."""
        await self.send(
            command=RoborockB01Q7Methods.SET_ROOM_CLEAN,
            params={
                "clean_type": CleanTaskTypeMapping.ALL.code,
                "ctrl_value": SCDeviceCleanParam.START.code,
                "room_ids": [],
            },
        )

    async def pause_clean(self) -> None:
        """Pause cleaning."""
        await self.send(
            command=RoborockB01Q7Methods.SET_ROOM_CLEAN,
            params={
                "clean_type": CleanTaskTypeMapping.ALL.code,
                "ctrl_value": SCDeviceCleanParam.PAUSE.code,
                "room_ids": [],
            },
        )

    async def stop_clean(self) -> None:
        """Stop cleaning."""
        await self.send(
            command=RoborockB01Q7Methods.SET_ROOM_CLEAN,
            params={
                "clean_type": CleanTaskTypeMapping.ALL.code,
                "ctrl_value": SCDeviceCleanParam.STOP.code,
                "room_ids": [],
            },
        )

    async def return_to_dock(self) -> None:
        """Return to dock."""
        await self.send(
            command=RoborockB01Q7Methods.START_RECHARGE,
            params={},
        )

    async def find_me(self) -> None:
        """Locate the robot."""
        await self.send(
            command=RoborockB01Q7Methods.FIND_DEVICE,
            params={},
        )

    async def send(self, command: CommandType, params: ParamsType) -> Any:
        """Send a command to the device."""
        return await send_decoded_command(
            self._channel,
            Q7RequestMessage(dps=10000, command=command, params=params),
        )


def create(channel: MqttChannel) -> Q7PropertiesApi:
    """Create traits for B01 devices."""
    return Q7PropertiesApi(channel)



"""Create traits for V1 devices.

Traits are modular components that encapsulate specific features of a Roborock
device. This module provides a factory function to create and initialize the
appropriate traits for V1 devices based on their capabilities.

Using Traits
------------
Traits are accessed via the `v1_properties` attribute on a device. Each trait
represents a specific capability, such as `status`, `consumables`, or `rooms`.

Traits serve two main purposes:
1.  **State**: Traits are dataclasses that hold the current state of the device
    feature. You can access attributes directly (e.g., `device.v1_properties.status.battery`).
2.  **Commands**: Traits provide methods to control the device. For example,
    `device.v1_properties.volume.set_volume()`.

Additionally, the `command` trait provides a generic way to send any command to the
device (e.g. `device.v1_properties.command.send("app_start")`). This is often used
for basic cleaning operations like starting, stopping, or docking the vacuum.

Most traits have a `refresh()` method that must be called to update their state
from the device. The state is not updated automatically in real-time unless
specifically implemented by the trait or via polling.

Adding New Traits
-----------------
When adding a new trait, the most common pattern is to subclass `V1TraitMixin`
and a `RoborockBase` dataclass. You must define a `command` class variable that
specifies the `RoborockCommand` used to fetch the trait data from the device.
See `common.py` for more details on common patterns used across traits.

There are some additional decorators in `common.py` that can be used to specify which
RPC channel to use for the trait (standard, MQTT/cloud, or map-specific).

  - `@common.mqtt_rpc_channel` - Use the MQTT RPC channel for this trait.
  - `@common.map_rpc_channel` - Use the map RPC channel for this trait.

There are also some attributes that specify device feature dependencies for
optional traits:

    - `requires_feature` - The string name of the device feature that must be supported
        for this trait to be enabled. See `DeviceFeaturesTrait` for a list of
        available features.
    - `requires_dock_type` - If set, this is a function that accepts a `RoborockDockTypeCode`
        and returns a boolean indicating whether the trait is supported for that dock type.

Additionally, DeviceFeaturesTrait has a method `is_field_supported` that is used to
check individual trait field values. This is a more fine grained version to allow
optional fields in a dataclass, vs the above feature checks that apply to an entire
trait. The `requires_schema_code` field metadata attribute is a string of the schema
code in HomeDataProduct Schema that is required for the field to be supported.
"""

import logging
from dataclasses import dataclass, field, fields
from typing import Any, get_args

from roborock.data.containers import HomeData, HomeDataProduct, RoborockBase
from roborock.data.v1.v1_code_mappings import RoborockDockTypeCode
from roborock.devices.cache import DeviceCache
from roborock.devices.traits import Trait
from roborock.map.map_parser import MapParserConfig
from roborock.protocols.v1_protocol import V1RpcChannel
from roborock.web_api import UserWebApiClient

from . import (
    child_lock,
    clean_summary,
    command,
    common,
    consumeable,
    device_features,
    do_not_disturb,
    dust_collection_mode,
    flow_led_status,
    home,
    led_status,
    map_content,
    maps,
    network_info,
    rooms,
    routines,
    smart_wash_params,
    status,
    valley_electricity_timer,
    volume,
    wash_towel_mode,
)
from .child_lock import ChildLockTrait
from .clean_summary import CleanSummaryTrait
from .command import CommandTrait
from .common import V1TraitMixin
from .consumeable import ConsumableTrait
from .device_features import DeviceFeaturesTrait
from .do_not_disturb import DoNotDisturbTrait
from .dust_collection_mode import DustCollectionModeTrait
from .flow_led_status import FlowLedStatusTrait
from .home import HomeTrait
from .led_status import LedStatusTrait
from .map_content import MapContentTrait
from .maps import MapsTrait
from .network_info import NetworkInfoTrait
from .rooms import RoomsTrait
from .routines import RoutinesTrait
from .smart_wash_params import SmartWashParamsTrait
from .status import StatusTrait
from .valley_electricity_timer import ValleyElectricityTimerTrait
from .volume import SoundVolumeTrait
from .wash_towel_mode import WashTowelModeTrait

_LOGGER = logging.getLogger(__name__)

__all__ = [
    "PropertiesApi",
    "child_lock",
    "clean_summary",
    "command",
    "common",
    "consumeable",
    "device_features",
    "do_not_disturb",
    "dust_collection_mode",
    "flow_led_status",
    "home",
    "led_status",
    "map_content",
    "maps",
    "network_info",
    "rooms",
    "routines",
    "smart_wash_params",
    "status",
    "valley_electricity_timer",
    "volume",
    "wash_towel_mode",
]


@dataclass
class PropertiesApi(Trait):
    """Common properties for V1 devices.

    This class holds all the traits that are common across all V1 devices.
    """

    # All v1 devices have these traits
    status: StatusTrait
    command: CommandTrait
    dnd: DoNotDisturbTrait
    clean_summary: CleanSummaryTrait
    sound_volume: SoundVolumeTrait
    rooms: RoomsTrait
    maps: MapsTrait
    map_content: MapContentTrait
    consumables: ConsumableTrait
    home: HomeTrait
    device_features: DeviceFeaturesTrait
    network_info: NetworkInfoTrait
    routines: RoutinesTrait

    # Optional features that may not be supported on all devices
    child_lock: ChildLockTrait | None = None
    led_status: LedStatusTrait | None = None
    flow_led_status: FlowLedStatusTrait | None = None
    valley_electricity_timer: ValleyElectricityTimerTrait | None = None
    dust_collection_mode: DustCollectionModeTrait | None = None
    wash_towel_mode: WashTowelModeTrait | None = None
    smart_wash_params: SmartWashParamsTrait | None = None

    def __init__(
        self,
        device_uid: str,
        product: HomeDataProduct,
        home_data: HomeData,
        rpc_channel: V1RpcChannel,
        mqtt_rpc_channel: V1RpcChannel,
        map_rpc_channel: V1RpcChannel,
        web_api: UserWebApiClient,
        device_cache: DeviceCache,
        map_parser_config: MapParserConfig | None = None,
        region: str | None = None,
    ) -> None:
        """Initialize the V1TraitProps."""
        self._device_uid = device_uid
        self._rpc_channel = rpc_channel
        self._mqtt_rpc_channel = mqtt_rpc_channel
        self._map_rpc_channel = map_rpc_channel
        self._web_api = web_api
        self._device_cache = device_cache
        self._region = region

        self.device_features = DeviceFeaturesTrait(product, self._device_cache)
        self.status = StatusTrait(self.device_features, region=self._region)
        self.consumables = ConsumableTrait()
        self.rooms = RoomsTrait(home_data, web_api)
        self.maps = MapsTrait(self.status)
        self.map_content = MapContentTrait(map_parser_config)
        self.home = HomeTrait(self.status, self.maps, self.map_content, self.rooms, self._device_cache)
        self.network_info = NetworkInfoTrait(device_uid, self._device_cache)
        self.routines = RoutinesTrait(device_uid, web_api)

        # Dynamically create any traits that need to be populated
        for item in fields(self):
            if (trait := getattr(self, item.name, None)) is None:
                # We exclude optional features and them via discover_features
                if (union_args := get_args(item.type)) is None or len(union_args) > 0:
                    continue
                _LOGGER.debug("Trait '%s' is supported, initializing", item.name)
                if not callable(item.type):
                    continue
                trait = item.type()
                setattr(self, item.name, trait)
            # This is a hack to allow setting the rpc_channel on all traits. This is
            # used so we can preserve the dataclass behavior when the values in the
            # traits are updated, but still want to allow them to have a reference
            # to the rpc channel for sending commands.
            trait._rpc_channel = self._get_rpc_channel(trait)

    def _get_rpc_channel(self, trait: V1TraitMixin) -> V1RpcChannel:
        # The decorator `@common.mqtt_rpc_channel` means that the trait needs
        # to use the mqtt_rpc_channel (cloud only) instead of the rpc_channel (adaptive)
        if hasattr(trait, "mqtt_rpc_channel"):
            return self._mqtt_rpc_channel
        elif hasattr(trait, "map_rpc_channel"):
            return self._map_rpc_channel
        else:
            return self._rpc_channel

    async def discover_features(self) -> None:
        """Populate any supported traits that were not initialized in __init__."""
        _LOGGER.debug("Starting optional trait discovery")
        await self.device_features.refresh()
        # Dock type also acts like a device feature for some traits.
        dock_type = await self._dock_type()

        # Dynamically create any traits that need to be populated
        for item in fields(self):
            if (trait := getattr(self, item.name, None)) is not None:
                continue
            if (union_args := get_args(item.type)) is None:
                raise ValueError(f"Unexpected non-union type for trait {item.name}: {item.type}")
            if len(union_args) != 2 or type(None) not in union_args:
                raise ValueError(f"Unexpected non-optional type for trait {item.name}: {item.type}")

            # Union args may not be in declared order
            item_type = union_args[0] if union_args[1] is type(None) else union_args[1]
            if not self._is_supported(item_type, item.name, dock_type):
                _LOGGER.debug("Trait '%s' not supported, skipping", item.name)
                continue
            _LOGGER.debug("Trait '%s' is supported, initializing", item.name)
            trait = item_type()
            setattr(self, item.name, trait)
            trait._rpc_channel = self._get_rpc_channel(trait)

    def _is_supported(self, trait_type: type[V1TraitMixin], name: str, dock_type: RoborockDockTypeCode) -> bool:
        """Check if a trait is supported by the device."""

        if (requires_dock_type := getattr(trait_type, "requires_dock_type", None)) is not None:
            return requires_dock_type(dock_type)

        if (feature_name := getattr(trait_type, "requires_feature", None)) is None:
            _LOGGER.debug("Optional trait missing 'requires_feature' attribute %s, skipping", name)
            return False
        if (is_supported := getattr(self.device_features, feature_name)) is None:
            raise ValueError(f"Device feature '{feature_name}' on trait '{name}' is unknown")
        return is_supported

    async def _dock_type(self) -> RoborockDockTypeCode:
        """Get the dock type from the status trait or cache."""
        dock_type = await self._get_cached_trait_data("dock_type")
        if dock_type is not None:
            _LOGGER.debug("Using cached dock type: %s", dock_type)
            try:
                return RoborockDockTypeCode(dock_type)
            except ValueError:
                _LOGGER.debug("Cached dock type %s is invalid, refreshing", dock_type)

        _LOGGER.debug("Starting dock type discovery")
        await self.status.refresh()
        _LOGGER.debug("Fetched dock type: %s", self.status.dock_type)
        if self.status.dock_type is None:
            # Explicitly set so we reuse cached value next type
            dock_type = RoborockDockTypeCode.no_dock
        else:
            dock_type = self.status.dock_type
        await self._set_cached_trait_data("dock_type", dock_type)
        return dock_type

    async def _get_cached_trait_data(self, name: str) -> Any:
        """Get the dock type from the status trait or cache."""
        cache_data = await self._device_cache.get()
        if cache_data.trait_data is None:
            cache_data.trait_data = {}
        _LOGGER.debug("Cached trait data: %s", cache_data.trait_data)
        return cache_data.trait_data.get(name)

    async def _set_cached_trait_data(self, name: str, value: Any) -> None:
        """Set trait-specific cached data."""
        cache_data = await self._device_cache.get()
        if cache_data.trait_data is None:
            cache_data.trait_data = {}
        cache_data.trait_data[name] = value
        _LOGGER.debug("Updating cached trait data: %s", cache_data.trait_data)
        await self._device_cache.set(cache_data)

    def as_dict(self) -> dict[str, Any]:
        """Return the trait data as a dictionary."""
        result: dict[str, Any] = {}
        for item in fields(self):
            trait = getattr(self, item.name, None)
            if trait is None or not isinstance(trait, RoborockBase):
                continue
            data = trait.as_dict()
            if data:  # Don't omit unset traits
                result[item.name] = data
        return result


def create(
    device_uid: str,
    product: HomeDataProduct,
    home_data: HomeData,
    rpc_channel: V1RpcChannel,
    mqtt_rpc_channel: V1RpcChannel,
    map_rpc_channel: V1RpcChannel,
    web_api: UserWebApiClient,
    device_cache: DeviceCache,
    map_parser_config: MapParserConfig | None = None,
    region: str | None = None,
) -> PropertiesApi:
    """Create traits for V1 devices."""
    return PropertiesApi(
        device_uid,
        product,
        home_data,
        rpc_channel,
        mqtt_rpc_channel,
        map_rpc_channel,
        web_api,
        device_cache,
        map_parser_config,
        region=region,
    )



"""Traits for Q10 B01 devices."""

import asyncio
import logging
from typing import Any

from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP
from roborock.devices.rpc.b01_q10_channel import stream_decoded_responses
from roborock.devices.traits import Trait
from roborock.devices.transport.mqtt_channel import MqttChannel

from .command import CommandTrait
from .status import StatusTrait
from .vacuum import VacuumTrait

__all__ = [
    "Q10PropertiesApi",
]

_LOGGER = logging.getLogger(__name__)


class Q10PropertiesApi(Trait):
    """API for interacting with B01 devices."""

    command: CommandTrait
    """Trait for sending commands to Q10 devices."""

    status: StatusTrait
    """Trait for managing the status of Q10 devices."""

    vacuum: VacuumTrait
    """Trait for sending vacuum related commands to Q10 devices."""

    def __init__(self, channel: MqttChannel) -> None:
        """Initialize the B01Props API."""
        self._channel = channel
        self.command = CommandTrait(channel)
        self.vacuum = VacuumTrait(self.command)
        self.status = StatusTrait()
        self._subscribe_task: asyncio.Task[None] | None = None

    async def start(self) -> None:
        """Start any necessary subscriptions for the trait."""
        self._subscribe_task = asyncio.create_task(self._subscribe_loop())

    async def close(self) -> None:
        """Close any resources held by the trait."""
        if self._subscribe_task is not None:
            self._subscribe_task.cancel()
            try:
                await self._subscribe_task
            except asyncio.CancelledError:
                pass  # ignore cancellation errors
            self._subscribe_task = None

    async def refresh(self) -> None:
        """Refresh all traits."""
        # Sending the REQUEST_DPS will cause the device to send all DPS values
        # to the device. Updates will be received by the subscribe loop below.
        await self.command.send(B01_Q10_DP.REQUEST_DPS, params={})

    async def _subscribe_loop(self) -> None:
        """Persistent loop to listen for status updates."""
        async for decoded_dps in stream_decoded_responses(self._channel):
            _LOGGER.debug("Received Q10 status update: %s", decoded_dps)

            # Notify all traits about a new message and each trait will
            # only update what fields that it is responsible for.
            # More traits can be added here below.
            self.status.update_from_dps(decoded_dps)


def create(channel: MqttChannel) -> Q10PropertiesApi:
    """Create traits for B01 devices."""
    return Q10PropertiesApi(channel)



"""Data container classes for Q10 B01 devices.

Many of these classes use the `field(metadata={"dps": ...})` convention to map
dataclass fields to device Data Points (DPS). This metadata is utilized by the
`update_from_dps` helper in `roborock.devices.traits.b01.q10.common` to
automatically update objects from raw device responses.
"""

from dataclasses import dataclass, field

from ..containers import RoborockBase
from .b01_q10_code_mappings import (
    B01_Q10_DP,
    YXBackType,
    YXDeviceCleanTask,
    YXDeviceState,
    YXDeviceWorkMode,
    YXFanLevel,
    YXWaterLevel,
)


@dataclass
class dpCleanRecord(RoborockBase):
    op: str
    result: int
    id: str
    data: list


@dataclass
class dpMultiMap(RoborockBase):
    op: str
    result: int
    data: list


@dataclass
class dpGetCarpet(RoborockBase):
    op: str
    result: int
    data: str


@dataclass
class dpSelfIdentifyingCarpet(RoborockBase):
    op: str
    result: int
    data: str


@dataclass
class dpNetInfo(RoborockBase):
    wifiName: str
    ipAdress: str
    mac: str
    signal: int


@dataclass
class dpNotDisturbExpand(RoborockBase):
    disturb_dust_enable: int
    disturb_light: int
    disturb_resume_clean: int
    disturb_voice: int


@dataclass
class dpCurrentCleanRoomIds(RoborockBase):
    room_id_list: list


@dataclass
class dpVoiceVersion(RoborockBase):
    version: int


@dataclass
class dpTimeZone(RoborockBase):
    timeZoneCity: str
    timeZoneSec: int


@dataclass
class Q10Status(RoborockBase):
    """Status for Q10 devices.

    Fields are mapped to DPS values using metadata. Objects of this class can be
    automatically updated using the `update_from_dps` helper.
    """

    clean_time: int | None = field(default=None, metadata={"dps": B01_Q10_DP.CLEAN_TIME})
    clean_area: int | None = field(default=None, metadata={"dps": B01_Q10_DP.CLEAN_AREA})
    battery: int | None = field(default=None, metadata={"dps": B01_Q10_DP.BATTERY})
    status: YXDeviceState | None = field(default=None, metadata={"dps": B01_Q10_DP.STATUS})
    fan_level: YXFanLevel | None = field(default=None, metadata={"dps": B01_Q10_DP.FAN_LEVEL})
    water_level: YXWaterLevel | None = field(default=None, metadata={"dps": B01_Q10_DP.WATER_LEVEL})
    clean_count: int | None = field(default=None, metadata={"dps": B01_Q10_DP.CLEAN_COUNT})
    clean_mode: YXDeviceWorkMode | None = field(default=None, metadata={"dps": B01_Q10_DP.CLEAN_MODE})
    clean_task_type: YXDeviceCleanTask | None = field(default=None, metadata={"dps": B01_Q10_DP.CLEAN_TASK_TYPE})
    back_type: YXBackType | None = field(default=None, metadata={"dps": B01_Q10_DP.BACK_TYPE})
    cleaning_progress: int | None = field(default=None, metadata={"dps": B01_Q10_DP.CLEANING_PROGRESS})



"""Module for handling network connections to Roborock devices.

This is used internally by the device manager for creating connections to
Roborock devices. These modules contain common code, not specific to a
particular device or application level protocol.
"""

__all__: list[str] = []



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum, StrEnum

from .data import (
    CleanRecord,
    CleanSummary,
    Consumable,
    DustCollectionMode,
    RoborockBase,
    SmartWashParams,
    Status,
    WashTowelMode,
)


class RoborockCommand(str, Enum):
    """Enum of all known Roborock V1 protocol commands.

    These commands can be sent to a device using the `CommandTrait`.
    For example: `device.v1_properties.command.send(RoborockCommand.APP_START)`.
    """

    ADD_MOP_TEMPLATE_PARAMS = "add_mop_template_params"
    APP_AMETHYST_SELF_CHECK = "app_amethyst_self_check"
    APP_CHARGE = "app_charge"
    APP_DELETE_WIFI = "app_delete_wifi"
    APP_GET_AMETHYST_STATUS = "app_get_amethyst_status"
    APP_GET_CARPET_DEEP_CLEAN_STATUS = "app_get_carpet_deep_clean_status"
    APP_GET_CLEAN_ESTIMATE_INFO = "app_get_clean_estimate_info"
    APP_GET_DRYER_SETTING = "app_get_dryer_setting"
    APP_GET_INIT_STATUS = "app_get_init_status"
    APP_GET_LOCALE = "app_get_locale"
    APP_GET_WIFI_LIST = "app_get_wifi_list"
    APP_GOTO_TARGET = "app_goto_target"
    APP_KEEP_EASTER_EGG = "app_keep_easter_egg"
    APP_PAUSE = "app_pause"
    APP_RC_END = "app_rc_end"
    APP_RC_MOVE = "app_rc_move"
    APP_RC_START = "app_rc_start"
    APP_RC_STOP = "app_rc_stop"
    APP_RESUME_BUILD_MAP = "app_resume_build_map"
    APP_RESUME_PATROL = "app_resume_patrol"
    APP_SEGMENT_CLEAN = "app_segment_clean"
    APP_SET_AMETHYST_STATUS = "app_set_amethyst_status"
    APP_SET_CARPET_DEEP_CLEAN_STATUS = "app_set_carpet_deep_clean_status"
    APP_SET_CROSS_CARPET_CLEANING_STATUS = "app_set_cross_carpet_cleaning_status"
    APP_SET_DOOR_SILL_BLOCKS = "app_set_door_sill_blocks"
    APP_SET_DIRTY_REPLENISH_CLEAN_STATUS = "app_set_dirty_replenish_clean_status"
    APP_SET_DRYER_SETTING = "app_set_dryer_setting"
    APP_SET_DRYER_STATUS = "app_set_dryer_status"
    APP_SET_DYNAMIC_CONFIG = "app_set_dynamic_config"
    APP_SET_IGNORE_STUCK_POINT = "app_set_ignore_stuck_point"
    APP_SET_SMART_CLIFF_FORBIDDEN = "app_set_smart_cliff_forbidden"
    APP_SET_SMART_DOOR_SILL = "app_set_smart_door_sill"
    APP_SPOT = "app_spot"
    APP_START = "app_start"
    APP_START_BUILD_MAP = "app_start_build_map"
    APP_START_COLLECT_DUST = "app_start_collect_dust"
    APP_START_EASTER_EGG = "app_start_easter_egg"
    APP_START_PATROL = "app_start_patrol"
    APP_START_PET_PATROL = "app_start_pet_patrol"
    APP_START_WASH = "app_start_wash"
    APP_STAT = "app_stat"
    APP_STOP = "app_stop"
    APP_STOP_COLLECT_DUST = "app_stop_collect_dust"
    APP_STOP_WASH = "app_stop_wash"
    APP_UPDATE_UNSAVE_MAP = "app_update_unsave_map"
    APP_WAKEUP_ROBOT = "app_wakeup_robot"
    APP_ZONED_CLEAN = "app_zoned_clean"
    CHANGE_SOUND_VOLUME = "change_sound_volume"
    CHECK_HOMESEC_PASSWORD = "check_homesec_password"
    CLOSE_DND_TIMER = "close_dnd_timer"
    CLOSE_VALLEY_ELECTRICITY_TIMER = "close_valley_electricity_timer"
    DEL_CLEAN_RECORD = "del_clean_record"
    DEL_CLEAN_RECORD_MAP_V2 = "del_clean_record_map_v2"
    DEL_MAP = "del_map"
    DEL_MOP_TEMPLATE_PARAMS = "del_mop_template_params"
    DEL_SERVER_TIMER = "del_server_timer"
    DEL_TIMER = "del_timer"
    DNLD_INSTALL_SOUND = "dnld_install_sound"
    ENABLE_HOMESEC_VOICE = "enable_homesec_voice"
    ENABLE_LOG_UPLOAD = "enable_log_upload"
    END_EDIT_MAP = "end_edit_map"
    FIND_ME = "find_me"
    GET_AUTO_DELIVERY_CLEANING_FLUID = "get_auto_delivery_cleaning_fluid"
    GET_CAMERA_STATUS = "get_camera_status"
    GET_CARPET_CLEAN_MODE = "get_carpet_clean_mode"
    GET_CARPET_MODE = "get_carpet_mode"
    GET_CHILD_LOCK_STATUS = "get_child_lock_status"
    GET_CLEAN_FOLLOW_GROUND_MATERIAL_STATUS = "get_clean_follow_ground_material_status"
    GET_CLEAN_MOTOR_MODE = "get_clean_motor_mode"
    GET_CLEAN_RECORD = "get_clean_record"
    GET_CLEAN_RECORD_MAP = "get_clean_record_map"
    GET_CLEAN_SEQUENCE = "get_clean_sequence"
    GET_CLEAN_SUMMARY = "get_clean_summary"
    GET_COLLISION_AVOID_STATUS = "get_collision_avoid_status"
    GET_CONSUMABLE = "get_consumable"
    GET_CURRENT_SOUND = "get_current_sound"
    GET_CUSTOM_MODE = "get_custom_mode"
    GET_CUSTOMIZE_CLEAN_MODE = "get_customize_clean_mode"
    GET_DEVICE_ICE = "get_device_ice"
    GET_DEVICE_SDP = "get_device_sdp"
    GET_DND_TIMER = "get_dnd_timer"
    GET_DOCK_INFO = "get_dock_info"
    GET_DUST_COLLECTION_MODE = "get_dust_collection_mode"
    GET_DUST_COLLECTION_SWITCH_STATUS = "get_dust_collection_switch_status"
    GET_DYNAMIC_DATA = "get_dynamic_data"
    GET_DYNAMIC_MAP_DIFF = "get_dynamic_map_diff"
    GET_FAN_MOTOR_WORK_TIMEOUT = "get_fan_motor_work_timeout"
    GET_FLOW_LED_STATUS = "get_flow_led_status"
    GET_FRESH_MAP = "get_fresh_map"
    GET_FW_FEATURES = "get_fw_features"
    GET_HOMESEC_CONNECT_STATUS = "get_homesec_connect_status"
    GET_IDENTIFY_FURNITURE_STATUS = "get_identify_furniture_status"
    GET_IDENTIFY_GROUND_MATERIAL_STATUS = "get_identify_ground_material_status"
    GET_LED_STATUS = "get_led_status"
    GET_LOG_UPLOAD_STATUS = "get_log_upload_status"
    GET_MAP = "get_map"
    GET_MAP_BEAUTIFICATION_STATUS = "get_map_beautification_status"
    GET_MAP_STATUS = "get_map_status"
    GET_MAP_V1 = "get_map_v1"
    GET_MAP_V2 = "get_map_v2"
    GET_MAP_CALIBRATION = "get_map_calibration"  # Custom command
    GET_MOP_MOTOR_STATUS = "get_mop_motor_status"
    GET_MOP_TEMPLATE_PARAMS_BY_ID = "get_mop_template_params_by_id"
    GET_MOP_TEMPLATE_PARAMS_SUMMARY = "get_mop_template_params_summary"
    GET_MULTI_MAP = "get_multi_map"
    GET_MULTI_MAPS_LIST = "get_multi_maps_list"
    GET_NETWORK_INFO = "get_network_info"
    GET_OFFLINE_MAP_STATUS = "get_offline_map_status"
    GET_PERSIST = "get_persist_map"
    GET_PROP = "get_prop"
    GET_RANDOM_PKEY = "get_random_pkey"
    GET_RECOVER_MAP = "get_recover_map"
    GET_RECOVER_MAPS = "get_recover_maps"
    GET_ROOM_MAPPING = "get_room_mapping"
    GET_SCENES_VALID_TIDS = "get_scenes_valid_tids"
    GET_SEGMENT_STATUS = "get_segment_status"
    GET_SERIAL_NUMBER = "get_serial_number"
    GET_SERVER_TIMER = "get_server_timer"
    GET_SMART_WASH_PARAMS = "get_smart_wash_params"
    GET_SOUND_PROGRESS = "get_sound_progress"
    GET_SOUND_VOLUME = "get_sound_volume"
    GET_STATUS = "get_status"
    GET_TESTID = "get_testid"
    GET_TIMER = "get_timer"
    GET_TIMER_DETAIL = "get_timer_detail"
    GET_TIMER_SUMMARY = "get_timer_summary"
    GET_TIMEZONE = "get_timezone"
    GET_TURN_SERVER = "get_turn_server"
    GET_VALLEY_ELECTRICITY_TIMER = "get_valley_electricity_timer"
    GET_WASH_DEBUG_PARAMS = "get_wash_debug_params"
    GET_WASH_TOWEL_MODE = "get_wash_towel_mode"
    GET_WASH_TOWEL_PARAMS = "get_wash_towel_params"
    GET_WATER_BOX_CUSTOM_MODE = "get_water_box_custom_mode"
    LOAD_MULTI_MAP = "load_multi_map"
    MANUAL_BAK_MAP = "manual_bak_map"
    MANUAL_SEGMENT_MAP = "manual_segment_map"
    MERGE_SEGMENT = "merge_segment"
    MOP_MODE = "mop_mode"
    MOP_TEMPLATE_ID = "mop_template_id"
    NAME_MULTI_MAP = "name_multi_map"
    NAME_SEGMENT = "name_segment"
    PLAY_AUDIO = "play_audio"
    RECOVER_MAP = "recover_map"
    RECOVER_MULTI_MAP = "recover_multi_map"
    RESET_CONSUMABLE = "reset_consumable"
    RESET_HOMESEC_PASSWORD = "reset_homesec_password"
    RESET_MAP = "reset_map"
    RESOLVE_ERROR = "resolve_error"
    RESUME_SEGMENT_CLEAN = "resume_segment_clean"
    RESUME_ZONED_CLEAN = "resume_zoned_clean"
    RETRY_REQUEST = "retry_request"
    REUNION_SCENES = "reunion_scenes"
    SAVE_FURNITURES = "save_furnitures"
    SAVE_MAP = "save_map"
    SEND_ICE_TO_ROBOT = "send_ice_to_robot"
    SEND_SDP_TO_ROBOT = "send_sdp_to_robot"
    SET_AIRDRY_HOURS = "set_airdry_hours"
    SET_APP_TIMEZONE = "set_app_timezone"
    SET_AUTO_DELIVERY_CLEANING_FLUID = "set_auto_delivery_cleaning_fluid"
    SET_CAMERA_STATUS = "set_camera_status"
    SET_CARPET_AREA = "set_carpet_area"
    SET_CARPET_CLEAN_MODE = "set_carpet_clean_mode"
    SET_CARPET_MODE = "set_carpet_mode"
    SET_CHILD_LOCK_STATUS = "set_child_lock_status"
    SET_CLEAN_FOLLOW_GROUND_MATERIAL_STATUS = "set_clean_follow_ground_material_status"
    SET_CLEAN_MOTOR_MODE = "set_clean_motor_mode"
    SET_CLEAN_SEQUENCE = "set_clean_sequence"
    SET_CLEAN_REPEAT_TIMES = "set_clean_repeat_times"
    SET_COLLISION_AVOID_STATUS = "set_collision_avoid_status"
    SET_CUSTOM_MODE = "set_custom_mode"
    SET_CUSTOMIZE_CLEAN_MODE = "set_customize_clean_mode"
    SET_DND_TIMER = "set_dnd_timer"
    SET_DND_TIMER_ACTIONS = "set_dnd_timer_actions"
    SET_DUST_COLLECTION_MODE = "set_dust_collection_mode"
    SET_DUST_COLLECTION_SWITCH_STATUS = "set_dust_collection_switch_status"
    SET_FAN_MOTOR_WORK_TIMEOUT = "set_fan_motor_work_timeout"
    SET_FDS_ENDPOINT = "set_fds_endpoint"
    SET_FLOW_LED_STATUS = "set_flow_led_status"
    SET_HOMESEC_PASSWORD = "set_homesec_password"
    SET_IDENTIFY_FURNITURE_STATUS = "set_identify_furniture_status"
    SET_IDENTIFY_GROUND_MATERIAL_STATUS = "set_identify_ground_material_status"
    SET_IGNORE_CARPET_ZONE = "set_ignore_carpet_zone"
    SET_IGNORE_IDENTIFY_AREA = "set_ignore_identify_area"
    SET_LAB_STATUS = "set_lab_status"
    SET_LED_STATUS = "set_led_status"
    SET_MAP_BEAUTIFICATION_STATUS = "set_map_beautification_status"
    SET_MOP_MODE = "set_mop_mode"
    SET_MOP_MOTOR_STATUS = "set_mop_motor_status"
    SET_MOP_TEMPLATE_ID = "set_mop_template_id"
    SET_OFFLINE_MAP_STATUS = "set_offline_map_status"
    SET_SCENES_SEGMENTS = "set_scenes_segments"
    SET_SCENES_ZONES = "set_scenes_zones"
    SET_SEGMENT_GROUND_MATERIAL = "set_segment_ground_material"
    SET_SERVER_TIMER = "set_server_timer"
    SET_SMART_WASH_PARAMS = "set_smart_wash_params"
    SET_SWITCH_MOP_MODE = "set_switch_map_mode"
    SET_TIMER = "set_timer"
    SET_TIMEZONE = "set_timezone"
    SET_VALLEY_ELECTRICITY_TIMER = "set_valley_electricity_timer"
    SET_VOICE_CHAT_VOLUME = "set_voice_chat_volume"
    SET_WASH_DEBUG_PARAMS = "set_wash_debug_params"
    SET_WASH_TOWEL_MODE = "set_wash_towel_mode"
    SET_WASH_TOWEL_PARAMS = "set_wash_towel_params"
    SET_WATER_BOX_CUSTOM_MODE = "set_water_box_custom_mode"
    SET_WATER_BOX_DISTANCE_OFF = "set_water_box_distance_off"
    SORT_MOP_TEMPLATE_PARAMS = "sort_mop_template_params"
    SPLIT_SEGMENT = "split_segment"
    START_CAMERA_PREVIEW = "start_camera_preview"
    START_CLEAN = "start_clean"
    START_EDIT_MAP = "start_edit_map"
    START_VOICE_CHAT = "start_voice_chat"
    START_WASH_THEN_CHARGE = "start_wash_then_charge"
    STOP_CAMERA_PREVIEW = "stop_camera_preview"
    STOP_FAN_MOTOR_WORK = "stop_fan_motor_work"
    STOP_GOTO_TARGET = "stop_goto_target"
    STOP_SEGMENT_CLEAN = "stop_segment_clean"
    STOP_VOICE_CHAT = "stop_voice_chat"
    STOP_ZONED_CLEAN = "stop_zoned_clean"
    SWITCH_VIDEO_QUALITY = "switch_video_quality"
    SWITCH_WATER_MARK = "switch_water_mark"
    TEST_SOUND_VOLUME = "test_sound_volume"
    UPD_SERVER_TIMER = "upd_server_timer"
    UPD_TIMER = "upd_timer"
    UPDATE_DOCK = "update_dock"
    UPDATE_MOP_TEMPLATE_PARAMS = "update_mop_template_params"
    UPLOAD_DATA_FOR_DEBUG_MODE = "upload_data_for_debug_mode"
    UPLOAD_PHOTO = "upload_photo"
    USE_NEW_MAP = "use_new_map"
    USE_OLD_MAP = "use_old_map"
    USER_UPLOAD_LOG = "user_upload_log"
    SET_STRETCH_TAG_STATUS = "set_stretch_tag_status"
    GET_STRETCH_TAG_STATUS = "get_stretch_tag_status"
    SET_RIGHT_BRUSH_STRETCH_STATUS = "set_right_brush_stretch_status"
    GET_RIGHT_BRUSH_STRETCH_STATUS = "get_right_brush_stretch_status"
    SET_DIRTY_OBJECT_DETECT_STATUS = "set_dirty_object_detect_status"
    GET_DIRTY_OBJECT_DETECT_STATUS = "get_dirty_object_detect_status"
    SET_WASH_WATER_TEMPERATURE = "set_wash_water_temperature"
    GET_WASH_WATER_TEMPERATURE = "get_wash_water_temperature"
    APP_EMPTY_RINSE_TANK_WATER = "app_empty_rinse_tank_water"
    SET_PET_SUPPLIES_DEEP_CLEAN_STATUS = "set_pet_supplies_deep_clean_status"
    GET_PET_SUPPLIES_DEEP_CLEAN_STATUS = "get_pet_supplies_deep_clean_status"
    SET_AP_MIC_LED_STATUS = "set_ap_mic_led_status"
    GET_AP_MIC_LED_STATUS = "get_ap_mic_led_status"
    SET_HANDLE_LEAK_WATER_STATUS = "set_handle_leak_water_status"
    GET_HANDLE_LEAK_WATER_STATUS = "get_handle_leak_water_status"
    APP_IGNORE_DIRTY_OBJECTS = "app_ignore_dirty_objects"
    MATTER_GET_STATUS = "matter.get_status"
    MATTER_DNLD_KEY = "matter.dnld_key"
    MATTER_RESET = "matter.reset"
    SET_GAP_DEEP_CLEAN_STATUS = "set_gap_deep_clean_status"
    GET_GAP_DEEP_CLEAN_STATUS = "get_gap_deep_clean_status"
    APP_SET_ROBOT_SETTING = "app_set_robot_setting"
    APP_GET_ROBOT_SETTING = "app_get_robot_setting"


class RoborockB01Q7Methods(StrEnum):
    """Methods used by the Roborock Q7 model."""

    # NOTE: In the Q7 Hermes dump these appear as suffixes and are also used
    # with an "event." prefix at runtime (see `hermes/.../module_524.js`).
    ADD_CLEAN_FAILED_POST = "add_clean_failed.post"
    EVENT_ADD_CLEAN_FAILED_POST = "event.add_clean_failed.post"
    CLEAN_FINISH_POST = "clean_finish.post"
    EVENT_CLEAN_FINISH_POST = "event.clean_finish.post"
    EVENT_BUILD_MAP_FINISH_POST = "event.BuildMapFinish.post"
    EVENT_MAP_CHANGE_POST = "event.map_change.post"
    EVENT_WORK_APPOINT_CLEAN_FAILED_POST = "event.work_appoint_clean_failed.post"
    START_CLEAN_POST = "startClean.post"
    ADD_ORDER = "service.add_order"
    ADD_SWEEP_CLEAN = "service.add_sweep_clean"
    ARRANGE_ROOM = "service.arrange_room"
    DEL_MAP = "service.del_map"
    DEL_ORDER = "service.del_order"
    DEL_ORDERS = "service.del_orders"
    DELETE_RECORD_BY_URL = "service.delete_record_by_url"
    DOWNLOAD_VOICE_TYPE = "service.download_voice_type"
    ERASE_PREFERENCE = "service.erase_preference"
    FIND_DEVICE = "service.find_device"
    GET_ROOM_ORDER = "service.get_room_order"
    GET_VOICE_DOWNLOAD = "service.get_voice_download"
    HELLO_WIKKA = "service.hello_wikka"
    RENAME_MAP = "service.rename_map"
    RENAME_ROOM = "service.rename_room"
    RENAME_ROOMS = "service.rename_rooms"
    REPLACE_MAP = "service.replace_map"
    RESET_CONSUMABLE = "service.reset_consumable"
    SAVE_CARPET = "service.save_carpet"
    SAVE_RECOMMEND_FB = "service.save_recommend_fb"
    SAVE_SILL = "service.save_sill"
    SET_AREA_START = "service.set_area_start"
    SET_AREAS_START = "service.set_areas_start"
    SET_CUR_MAP = "service.set_cur_map"
    SET_DIRECTION = "service.set_direction"
    SET_GLOBAL_SORT = "service.set_global_sort"
    SET_MAP_HIDE = "service.set_map_hide"
    SET_MULTI_ROOM_MATERIAL = "service.set_multi_room_material"
    SET_POINT_CLEAN = "service.set_point_clean"
    SET_PREFERENCE = "service.set_preference"
    SET_PREFERENCE_TYPE = "service.set_preference_type"
    SET_QUIET_TIME = "service.set_quiet_time"
    SET_ROOM_CLEAN = "service.set_room_clean"
    SET_ROOM_ORDER = "service.set_room_order"
    SET_VIRTUAL_WALL = "service.set_virtual_wall"
    SET_ZONE_CLEAN = "service.set_zone_clean"
    SET_ZONE_POINTS = "service.set_zone_points"
    SPLIT_ROOM = "service.split_room"
    START_EXPLORE = "service.start_explore"
    START_POINT_CLEAN = "service.start_point_clean"
    START_RECHARGE = "service.start_recharge"
    STOP_RECHARGE = "service.stop_recharge"
    UPLOAD_BY_MAPID = "service.upload_by_mapid"
    UPLOAD_RECORD_BY_URL = "service.upload_record_by_url"
    GET_PROP = "prop.get"
    GET_MAP_LIST = "service.get_map_list"
    UPLOAD_BY_MAPTYPE = "service.upload_by_maptype"
    SET_PROP = "prop.set"
    GET_PREFERENCE = "service.get_preference"
    GET_RECORD_LIST = "service.get_record_list"
    GET_ORDER = "service.get_order"
    POST_PROP = "prop.post"


@dataclass
class DockSummary(RoborockBase):
    dust_collection_mode: DustCollectionMode | None = None
    wash_towel_mode: WashTowelMode | None = None
    smart_wash_params: SmartWashParams | None = None


@dataclass
class DeviceProp(RoborockBase):
    status: Status = field(default_factory=Status)
    clean_summary: CleanSummary = field(default_factory=CleanSummary)
    consumable: Consumable = field(default_factory=Consumable)
    last_clean_record: CleanRecord | None = None
    dock_summary: DockSummary | None = None
    dust_collection_mode_name: str | None = None

    def __post_init__(self) -> None:
        if (
            self.dock_summary
            and self.dock_summary.dust_collection_mode is not None
            and self.dock_summary.dust_collection_mode.mode is not None
        ):
            self.dust_collection_mode_name = self.dock_summary.dust_collection_mode.mode.name

    def update(self, device_prop: DeviceProp) -> None:
        if device_prop.status:
            self.status = device_prop.status
        if device_prop.clean_summary:
            self.clean_summary = device_prop.clean_summary
        if device_prop.consumable:
            self.consumable = device_prop.consumable
        if device_prop.last_clean_record:
            self.last_clean_record = device_prop.last_clean_record
        if device_prop.dock_summary:
            self.dock_summary = device_prop.dock_summary
        self.__post_init__()



"""Low-level interface for connections to Roborock devices."""

import logging
from collections.abc import Callable
from typing import Protocol

from roborock.roborock_message import RoborockMessage

_LOGGER = logging.getLogger(__name__)


class Channel(Protocol):
    """A generic channel for establishing a connection with a Roborock device.

    Individual channel implementations have their own methods for speaking to
    the device that hide some of the protocol specific complexity, but they
    are still specialized for the device type and protocol.
    """

    @property
    def is_connected(self) -> bool:
        """Return true if the channel is connected."""
        ...

    @property
    def is_local_connected(self) -> bool:
        """Return true if the channel is connected locally."""
        ...

    async def subscribe(self, callback: Callable[[RoborockMessage], None]) -> Callable[[], None]:
        """Subscribe to messages from the device."""
        ...



"""An MQTT session for sending and receiving messages.

See create_mqtt_session for a factory function to create an MQTT session.

This is a thin wrapper around the async MQTT client that handles dispatching messages
from a topic to a callback function, since the async MQTT client does not
support this out of the box. It also handles the authentication process and
receiving messages from the vacuum cleaner.
"""

import asyncio
import datetime
import logging
import ssl
from collections.abc import Callable
from contextlib import asynccontextmanager

import aiomqtt
from aiomqtt import MqttCodeError, MqttError, TLSParameters

from roborock.callbacks import CallbackMap
from roborock.diagnostics import Diagnostics, redact_topic_name

from .health_manager import HealthManager
from .session import MqttParams, MqttSession, MqttSessionException, MqttSessionUnauthorized

_LOGGER = logging.getLogger(__name__)
_MQTT_LOGGER = logging.getLogger(f"{__name__}.aiomqtt")

CLIENT_KEEPALIVE = datetime.timedelta(seconds=45)
TOPIC_KEEPALIVE = datetime.timedelta(seconds=60)

# Exponential backoff parameters
MIN_BACKOFF_INTERVAL = datetime.timedelta(seconds=10)
MAX_BACKOFF_INTERVAL = datetime.timedelta(hours=6)
BACKOFF_MULTIPLIER = 1.5


class MqttReasonCode:
    """MQTT Reason Codes used by Roborock devices.

    This is a subset of paho.mqtt.reasoncodes.ReasonCode where we would like
    different error handling behavior.
    """

    RC_ERROR_UNAUTHORIZED = 135


class RoborockMqttSession(MqttSession):
    """An MQTT session for sending and receiving messages.

    You can start a session invoking the start() method which will connect to
    the MQTT broker. A caller may subscribe to a topic, and the session keeps
    track of which callbacks to invoke for each topic.

    The client is run as a background task that will run until shutdown. Once
    connected, the client will wait for messages to be received in a loop. If
    the connection is lost, the client will be re-created and reconnected. There
    is backoff to avoid spamming the broker with connection attempts.

    Reconnect attempts are deferred while there are no active subscriptions,
    which avoids unnecessary reconnect churn for idle sessions. Reconnects
    resume as soon as a subscription is added again. The client automatically
    re-establishes any existing subscriptions when the connection returns.
    """

    def __init__(
        self,
        params: MqttParams,
        topic_idle_timeout: datetime.timedelta = TOPIC_KEEPALIVE,
    ):
        self._params = params
        self._reconnect_task: asyncio.Task[None] | None = None
        self._healthy = False
        self._stop = False
        self._backoff = MIN_BACKOFF_INTERVAL
        self._client: aiomqtt.Client | None = None
        self._client_subscribed_topics: set[str] = set()
        self._client_lock = asyncio.Lock()
        self._listeners: CallbackMap[str, bytes] = CallbackMap(_LOGGER)
        self._connection_task: asyncio.Task[None] | None = None
        self._topic_idle_timeout = topic_idle_timeout
        self._idle_timers: dict[str, asyncio.Task[None]] = {}
        self._diagnostics = params.diagnostics
        self._health_manager = HealthManager(self.restart)
        self._unauthorized_hook = params.unauthorized_hook

    @property
    def connected(self) -> bool:
        """True if the session is connected to the broker."""
        return self._healthy

    @property
    def health_manager(self) -> HealthManager:
        """Return the health manager for the session."""
        return self._health_manager

    async def start(self) -> None:
        """Start the MQTT session.

        This has special behavior for the first connection attempt where any
        failures are raised immediately. This is to allow the caller to
        handle the failure and retry if desired itself. Once connected,
        the session will retry connecting in the background.
        """
        self._diagnostics.increment("start_attempt")
        start_future: asyncio.Future[None] = asyncio.Future()
        loop = asyncio.get_event_loop()
        self._reconnect_task = loop.create_task(self._run_reconnect_loop(start_future))
        try:
            await start_future
        except MqttCodeError as err:
            self._diagnostics.increment(f"start_failure:{err.rc}")
            if err.rc == MqttReasonCode.RC_ERROR_UNAUTHORIZED:
                raise MqttSessionUnauthorized(f"Authorization error starting MQTT session: {err}") from err
            raise MqttSessionException(f"Error starting MQTT session: {err}") from err
        except MqttError as err:
            self._diagnostics.increment("start_failure:unknown")
            raise MqttSessionException(f"Error starting MQTT session: {err}") from err
        except Exception as err:
            self._diagnostics.increment("start_failure:uncaught")
            raise MqttSessionException(f"Unexpected error starting session: {err}") from err
        else:
            self._diagnostics.increment("start_success")
            _LOGGER.debug("MQTT session started successfully")

    async def close(self) -> None:
        """Cancels the MQTT loop and shutdown the client library."""
        self._diagnostics.increment("close")
        self._stop = True
        tasks = [task for task in [self._connection_task, self._reconnect_task, *self._idle_timers.values()] if task]
        self._connection_task = None
        self._reconnect_task = None
        self._idle_timers.clear()

        for task in tasks:
            task.cancel()
        try:
            await asyncio.gather(*tasks, return_exceptions=True)
        except asyncio.CancelledError:
            pass

        self._healthy = False

    async def restart(self) -> None:
        """Force the session to disconnect and reconnect.

        The active connection task will be cancelled and restarted in the background, retried by
        the reconnect loop. This is a no-op if there is no active connection.
        """
        _LOGGER.info("Forcing MQTT session restart")
        self._diagnostics.increment("restart")
        if self._connection_task:
            self._connection_task.cancel()
        else:
            _LOGGER.debug("No message loop task to cancel")

    async def _run_reconnect_loop(self, start_future: asyncio.Future[None] | None) -> None:
        """Run the MQTT loop."""
        _LOGGER.info("Starting MQTT session")
        self._diagnostics.increment("start_loop")
        while True:
            try:
                self._connection_task = asyncio.create_task(self._run_connection(start_future))
                await self._connection_task
            except asyncio.CancelledError:
                _LOGGER.debug("MQTT connection task cancelled")
            except Exception:
                # Exceptions are logged and handled in _run_connection.
                # There is a special case for exceptions on startup where we return
                # immediately. Otherwise, we let the reconnect loop retry with
                # backoff when the reconnect loop is active.
                if start_future and start_future.done() and start_future.exception():
                    return

            self._healthy = False
            start_future = None
            if self._stop:
                _LOGGER.debug("MQTT session closed, stopping retry loop")
                return
            if not self._client_subscribed_topics and not self._listeners.keys():
                _LOGGER.debug("MQTT session disconnected with no active subscriptions, deferring reconnect")
                self._diagnostics.increment("reconnect_deferred")
                while not self._stop and not self._client_subscribed_topics and not self._listeners.keys():
                    await asyncio.sleep(0.1)
                if self._stop:
                    _LOGGER.debug("MQTT session closed while waiting for active subscriptions")
                    return
                self._backoff = MIN_BACKOFF_INTERVAL
                continue
            _LOGGER.info("MQTT session disconnected, retrying in %s seconds", self._backoff.total_seconds())
            self._diagnostics.increment("reconnect_wait")
            await asyncio.sleep(self._backoff.total_seconds())
            self._backoff = min(self._backoff * BACKOFF_MULTIPLIER, MAX_BACKOFF_INTERVAL)

    async def _run_connection(self, start_future: asyncio.Future[None] | None) -> None:
        """Connect to the MQTT broker and listen for messages.

        This is the primary connection loop for the MQTT session that is
        long running and processes incoming messages. If the connection
        is lost, this method will exit.
        """
        try:
            with self._diagnostics.timer("connection"):
                async with self._mqtt_client(self._params) as client:
                    self._backoff = MIN_BACKOFF_INTERVAL
                    self._healthy = True
                    _LOGGER.info("MQTT Session connected.")
                    if start_future and not start_future.done():
                        start_future.set_result(None)

                    _LOGGER.debug("Processing MQTT messages")
                    async for message in client.messages:
                        _LOGGER.debug("Received message: %s", message)
                        with self._diagnostics.timer("dispatch_message"):
                            self._listeners(message.topic.value, message.payload)
        except MqttCodeError as err:
            self._diagnostics.increment(f"connect_failure:{err.rc}")
            if start_future and not start_future.done():
                _LOGGER.debug("MQTT error starting session: %s", err)
                start_future.set_exception(err)
            else:
                _LOGGER.debug("MQTT error: %s", err)
            if err.rc == MqttReasonCode.RC_ERROR_UNAUTHORIZED and self._unauthorized_hook:
                _LOGGER.info("MQTT unauthorized/rate-limit error received, setting backoff to maximum")
                self._unauthorized_hook()
                self._backoff = MAX_BACKOFF_INTERVAL
            raise
        except MqttError as err:
            self._diagnostics.increment("connect_failure:unknown")
            if start_future and not start_future.done():
                _LOGGER.info("MQTT error starting session: %s", err)
                start_future.set_exception(err)
            else:
                _LOGGER.info("MQTT error: %s", err)
            raise
        except Exception as err:
            self._diagnostics.increment("connect_failure:uncaught")
            # This error is thrown when the MQTT loop is cancelled
            # and the generator is not stopped.
            if "generator didn't stop" in str(err) or "generator didn't yield" in str(err):
                _LOGGER.debug("MQTT loop was cancelled")
                return
            if start_future and not start_future.done():
                _LOGGER.error("Uncaught error starting MQTT session: %s", err)
                start_future.set_exception(err)
            else:
                _LOGGER.exception("Uncaught error during MQTT session: %s", err)
            raise

    @asynccontextmanager
    async def _mqtt_client(self, params: MqttParams) -> aiomqtt.Client:
        """Connect to the MQTT broker and listen for messages."""
        _LOGGER.debug("Connecting to %s:%s for %s", params.host, params.port, params.username)
        tls_params = None
        if params.tls:
            tls_params = TLSParameters(cert_reqs=ssl.CERT_REQUIRED if params.verify_tls else ssl.CERT_NONE)
        try:
            async with aiomqtt.Client(
                hostname=params.host,
                port=params.port,
                username=params.username,
                password=params.password,
                keepalive=int(CLIENT_KEEPALIVE.total_seconds()),
                protocol=aiomqtt.ProtocolVersion.V5,
                tls_params=tls_params,
                timeout=params.timeout,
                logger=_MQTT_LOGGER,
            ) as client:
                _LOGGER.debug("Connected to MQTT broker")
                # Re-establish any existing subscriptions
                async with self._client_lock:
                    self._client = client
                    for topic in self._client_subscribed_topics:
                        self._diagnostics.increment("resubscribe")
                        _LOGGER.debug("Re-establishing subscription to topic %s", redact_topic_name(topic))
                        # TODO: If this fails it will break the whole connection. Make
                        # this retry again in the background with backoff.
                        await client.subscribe(topic)

                yield client
        finally:
            async with self._client_lock:
                self._client = None

    async def subscribe(self, topic: str, callback: Callable[[bytes], None]) -> Callable[[], None]:
        """Subscribe to messages on the specified topic and invoke the callback for new messages.

        The callback will be called with the message payload as a bytes object. The callback
        should not block since it runs in the async loop. It should not raise any exceptions.

        The returned callable unsubscribes from the topic when called, but will delay actual
        unsubscription for the idle timeout period. If a new subscription comes in during the
        timeout, the timer is cancelled and the subscription is reused.
        """
        _LOGGER.debug("Subscribing to topic %s", redact_topic_name(topic))

        # If there is an idle timer for this topic, cancel it (reuse subscription)
        if idle_timer := self._idle_timers.pop(topic, None):
            self._diagnostics.increment("unsubscribe_idle_cancel")
            idle_timer.cancel()
            _LOGGER.debug("Cancelled idle timer for topic %s (reused subscription)", redact_topic_name(topic))

        unsub = self._listeners.add_callback(topic, callback)

        async with self._client_lock:
            if topic not in self._client_subscribed_topics:
                self._client_subscribed_topics.add(topic)
                if self._client:
                    _LOGGER.debug("Establishing subscription to topic %s", topic)
                    try:
                        with self._diagnostics.timer("subscribe"):
                            await self._client.subscribe(topic)
                    except MqttError as err:
                        # Clean up the callback if subscription fails
                        unsub()
                        self._client_subscribed_topics.discard(topic)
                        raise MqttSessionException(f"Error subscribing to topic: {err}") from err
                else:
                    self._diagnostics.increment("subscribe_pending")
                    _LOGGER.debug("Client not connected, will establish subscription later")

        def schedule_unsubscribe() -> None:
            async def idle_unsubscribe():
                try:
                    await asyncio.sleep(self._topic_idle_timeout.total_seconds())
                    # Only unsubscribe if there are no callbacks left for this topic
                    if not self._listeners.get_callbacks(topic):
                        async with self._client_lock:
                            # Check again if we have listeners, in case a subscribe happened
                            # while we were waiting for the lock or after we popped the timer.
                            if self._listeners.get_callbacks(topic):
                                _LOGGER.debug("Skipping unsubscribe for %s, new listeners added", topic)
                                return

                            self._idle_timers.pop(topic, None)
                            self._client_subscribed_topics.discard(topic)

                            if self._client:
                                _LOGGER.debug("Idle timeout expired, unsubscribing from topic %s", topic)
                                try:
                                    await self._client.unsubscribe(topic)
                                except MqttError as err:
                                    _LOGGER.warning("Error unsubscribing from topic %s: %s", topic, err)
                except asyncio.CancelledError:
                    _LOGGER.debug("Idle unsubscribe for topic %s cancelled", topic)

            # Start the idle timer task
            task = asyncio.create_task(idle_unsubscribe())
            self._idle_timers[topic] = task

        def delayed_unsub():
            self._diagnostics.increment("unsubscribe")
            unsub()  # Remove the callback from CallbackMap
            # If no more callbacks for this topic, start idle timer
            if not self._listeners.get_callbacks(topic):
                self._diagnostics.increment("unsubscribe_idle_start")
                schedule_unsubscribe()
            else:
                _LOGGER.debug("Unsubscribing topic %s, still have active callbacks", topic)

        return delayed_unsub

    async def publish(self, topic: str, message: bytes) -> None:
        """Publish a message on the topic."""
        _LOGGER.debug("Sending message to topic %s: %s", topic, message)
        client: aiomqtt.Client
        async with self._client_lock:
            if self._client is None:
                raise MqttSessionException("Could not publish message, MQTT client not connected")
            client = self._client
        try:
            with self._diagnostics.timer("publish"):
                await client.publish(topic, message)
        except MqttError as err:
            raise MqttSessionException(f"Error publishing message: {err}") from err


class LazyMqttSession(MqttSession):
    """An MQTT session that is started on first attempt to subscribe.

    This is a wrapper around an existing MqttSession that will only start
    the underlying session when the first attempt to subscribe or publish
    is made.
    """

    def __init__(self, session: RoborockMqttSession, diagnostics: Diagnostics) -> None:
        """Initialize the lazy session with an existing session."""
        self._lock = asyncio.Lock()
        self._started = False
        self._session = session
        self._diagnostics = diagnostics

    @property
    def connected(self) -> bool:
        """True if the session is connected to the broker."""
        return self._session.connected

    @property
    def health_manager(self) -> HealthManager:
        """Return the health manager for the session."""
        return self._session.health_manager

    async def _maybe_start(self) -> None:
        """Start the MQTT session if not already started."""
        async with self._lock:
            if not self._started:
                self._diagnostics.increment("start")
                await self._session.start()
                self._started = True

    async def subscribe(self, device_id: str, callback: Callable[[bytes], None]) -> Callable[[], None]:
        """Invoke the callback when messages are received on the topic.

        The returned callable unsubscribes from the topic when called.
        """
        await self._maybe_start()
        return await self._session.subscribe(device_id, callback)

    async def publish(self, topic: str, message: bytes) -> None:
        """Publish a message on the specified topic.

        This will raise an exception if the message could not be sent.
        """
        await self._maybe_start()
        return await self._session.publish(topic, message)

    async def close(self) -> None:
        """Cancels the mqtt loop.

        This will close the underlying session and will not allow it to be
        restarted again.
        """
        await self._session.close()

    async def restart(self) -> None:
        """Force the session to disconnect and reconnect."""
        await self._session.restart()


async def create_mqtt_session(params: MqttParams) -> MqttSession:
    """Create an MQTT session.

    This function is a factory for creating an MQTT session. This will
    raise an exception if initial attempt to connect fails. Once connected,
    the session will retry connecting on failure in the background.
    """
    session = RoborockMqttSession(params)
    await session.start()
    return session


async def create_lazy_mqtt_session(params: MqttParams) -> MqttSession:
    """Create a lazy MQTT session.

    This function is a factory for creating an MQTT session that will
    only connect when the first attempt to subscribe or publish is made.
    """
    return LazyMqttSession(RoborockMqttSession(params), diagnostics=params.diagnostics.subkey("lazy_mqtt"))



"""Traits for B01 devices."""

from . import q7, q10
from .q7 import Q7PropertiesApi
from .q10 import Q10PropertiesApi

__all__ = [
    "Q7PropertiesApi",
    "Q10PropertiesApi",
    "q7",
    "q10",
]



from dataclasses import dataclass

from ..containers import RoborockBase


@dataclass
class DyadProductInfo(RoborockBase):
    sn: str
    ssid: str
    timezone: str
    posix_timezone: str
    ip: str
    mac: str
    oba: dict


@dataclass
class DyadSndState(RoborockBase):
    sid_in_use: int
    sid_version: int
    location: str
    bom: str
    language: str


@dataclass
class DyadOtaNfo(RoborockBase):
    mqttOtaData: dict



"""Traits for Q10 B01 devices."""

from roborock.data.b01_q10.b01_q10_code_mappings import (
    B01_Q10_DP,
    YXCleanType,
    YXFanLevel,
)

from .command import CommandTrait


class VacuumTrait:
    """Trait for sending vacuum commands.

    This is a wrapper around the CommandTrait for sending vacuum related
    commands to Q10 devices.
    """

    def __init__(self, command: CommandTrait) -> None:
        """Initialize the VacuumTrait."""
        self._command = command

    async def start_clean(self) -> None:
        """Start cleaning."""
        await self._command.send(
            command=B01_Q10_DP.START_CLEAN,
            # TODO: figure out other commands
            # 1 = start cleaning
            # 2 = "electoral" clean, also has "clean_parameters"
            # 4 = fast create map
            params={"cmd": 1},
        )

    async def pause_clean(self) -> None:
        """Pause cleaning."""
        await self._command.send(
            command=B01_Q10_DP.PAUSE,
            params={},
        )

    async def resume_clean(self) -> None:
        """Resume cleaning."""
        await self._command.send(
            command=B01_Q10_DP.RESUME,
            params={},
        )

    async def stop_clean(self) -> None:
        """Stop cleaning."""
        await self._command.send(
            command=B01_Q10_DP.STOP,
            params={},
        )

    async def return_to_dock(self) -> None:
        """Return to dock."""
        await self._command.send(
            command=B01_Q10_DP.START_DOCK_TASK,
            params={},
        )

    async def empty_dustbin(self) -> None:
        """Empty the dustbin at the dock."""
        await self._command.send(
            command=B01_Q10_DP.START_DOCK_TASK,
            params=2,  # 2 = dock task type for "empty dustbin"
        )

    async def set_clean_mode(self, mode: YXCleanType) -> None:
        """Set the cleaning mode (vacuum, mop, or both)."""
        await self._command.send(
            command=B01_Q10_DP.CLEAN_MODE,
            params=mode.code,
        )

    async def set_fan_level(self, level: YXFanLevel) -> None:
        """Set the fan suction level."""
        await self._command.send(
            command=B01_Q10_DP.FAN_LEVEL,
            params=level.code,
        )



"""
CLI 入口

子命令模式：
  roborock-mcp          → 默认启动 stdio server
  roborock-mcp auth     → 运行交互式认证
  roborock-mcp --help   → 显示帮助
"""

import argparse
import sys


def main():
    parser = argparse.ArgumentParser(
        prog="roborock-mcp",
        description="Roborock MCP Server — 让 Claude 控制石头扫地机",
        epilog="不带子命令时默认启动 MCP server（stdio 模式）",
    )
    subparsers = parser.add_subparsers(dest="command")
    subparsers.add_parser("auth", help="首次认证（交互式邮箱验证码）")

    args = parser.parse_args()

    if args.command == "auth":
        from .setup_auth import main as auth_main
        auth_main()
    elif args.command is None:
        # 默认行为：启动 server
        from .server import mcp
        mcp.run(transport="stdio")
    else:
        # 防御性编程：argparse 通常会拦截未知子命令
        parser.error(f"未知子命令: {args.command}")
        sys.exit(1)


if __name__ == "__main__":
    main()
